#####################################################
# This script is meant to regrid and dealiase the polarimetric spectra that were measured during the tripex-pol Campaign. 
#####################################################

import os
import time
import numpy as np
import pandas as pd
import xarray as xr
import glob as glob
from sys import argv
#from matplotlib import pyplot as plt
#own routines:
import process_Wband_functions as pro
#import plotting_routines as plot
debugging = False

dateStart = pd.to_datetime('20220201'); dateEnd = pd.to_datetime('20220228')
dateList = pd.date_range(dateStart, dateEnd,freq='D')
dataOutPath = '/project/meteo/work/L.Terzi/tripex_pol_scan/output/'


for dayIndex,date2proc in enumerate(dateList):
	dataPath = '/archive/meteo/external-obs/juelich/tripex-pol-scan/wband_scan/l0/{year}/{month}/{day}'.format(year=date2proc.strftime('%Y'),
		                                                                             month=date2proc.strftime('%m'),
		                                                                             day=date2proc.strftime('%d'))
	dataOutputPath = dataOutPath + '{year}/{month}/{day}/'.format(year=date2proc.strftime('%Y'),
				                                              month=date2proc.strftime('%m'),
				                                              day=date2proc.strftime('%d'))
	
	fileId = 'CEL__{dateStr}_??????_P06_CEL.LV0.nc'.format(dateStr=date2proc.strftime('%y%m%d'))
	filePath = os.path.join(dataPath, fileId)
	fileList=sorted(glob.glob(filePath)) #=sorted(glob.glob(filePath2)) #.extend(sorted(glob.glob(filePath2)))
	if fileList: #- only process if files are there
		if debugging==True:
			print(fileList)
			#quit()
		for f in fileList:
			print(f)
			#quit()
			file2save = f.split('/')[-1].split('.nc')[0]+'.dealiasedNew.nc'
			#quit()
			data = xr.open_dataset(f)
			 
			time = data.Time.values + data.Timems.values/1000.
			data = data.assign({'Time':time})
			data.Time.attrs['units']='Seconds since 01.01.2001 00:00:00'
			data = xr.decode_cf(data)
			# get rid of duplicate time values:
			_, index_time = np.unique(data['Time'], return_index=True)
			data = data.isel(Time=index_time)
			
			#-- now merge the chirps into one range, and while we are at it, I am going to recalculate the VSpec (noise doesnt need to be removed, this is already done automatically in he RPG software)
			# (the VSpec that is in this dataset for now is not really the VSpec, but a composite of H and V, and with the ChReVHSpec variable we can calculate the real VSpec.
			if debugging==True:
				print('now merging chirps')
			dataMerged = pro.mergeChirps(data,spec=True)
			#quit()
			dataMerged['sSNR_H'] = 10*np.log10(dataMerged['sSNR_H'])
			dataMerged['sSNR_V'] = 10*np.log10(dataMerged['sSNR_V'])
			dataMerged['HSpec'] = dataMerged['HSpec'].where(dataMerged['sSNR_H']>10)#>10**(-37/10))
			dataMerged['sZDR'] = 10*np.log10(dataMerged['HSpec']) - 10*np.log10(dataMerged['VSpec'])
			
			#- dealize the spectra:
			if debugging==True:
				print('now dealiasing')
				
			#dataDealized = pro.dealiase(dataMerged)
			dataDealized = pro.dealiazeParallel(dataMerged)
			if debugging==True:
				print('now done dealiasing, saving the file again')
			if not os.path.isdir(dataOutputPath):
					os.makedirs(dataOutputPath)
			dataDealized.to_netcdf('{path}{file}'.format(path=dataOutputPath,file=file2save))




from pathlib import Path
from pdb import set_trace
import matplotlib.pyplot as plt
import xarray as xr

from gfatpy.lidar.utils.utils import signal_to_rcs


def plot_eta_star_calib(
    d: xr.Dataset,
    wavelength: int = 532,
    output_dir: Path | str | None = None,
    telescope: str = "x",
    **kwargs,
) -> Path | None:
    """It plots the calibration factors of a specific channel (i.e., wavelength and telescope) from dataset load with `gfatpy.lidar.depolarization.calibration.eta_star_reader`. 

    Args:
    
        d (xr.Dataset): calibration factor dataset.
    """
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    elif output_dir is None:
        output_dir = Path.cwd()

    if not output_dir.exists() or not output_dir.is_dir():
        raise NotADirectoryError(f"{output_dir} not found.")

    try:
        channel_str = f"{wavelength}{telescope}"

        # FIXME: This search channels should be provisional and shoud be make by
        # passing a channels array with the eta star dataset
        variables: list[str] = list(d.variables.keys())  # type: ignore
        search_channels = filter(lambda v: v.startswith("eta_star_profile_"), variables)

        _, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(10, 10), sharey=True) # type: ignore
        y_lim = kwargs.get("y_lim", (0, 5000))

        # Title
        date_str, hour_str = d.attrs['calibration_datetime'].split('_')        
        system = d.attrs['system'].upper()
        fig_title_1 = f"{system} calibration depolarization analysis | {date_str} - {hour_str} UTC | {channel_str}"
        cal_height_an_min, cal_height_an_max = d[f'eta_star_mean_{channel_str}a'].attrs['min_range_m'], d[f'eta_star_mean_{channel_str}a'].attrs['max_range_m']
        cal_height_pc_min, cal_height_pc_max = d[f'eta_star_mean_{channel_str}p'].attrs['min_range_m'], d[f'eta_star_mean_{channel_str}p'].attrs['max_range_m']
        eta_star_an = d[f"eta_star_mean_{channel_str}a"].values.item()
        eta_star_pc = d[f"eta_star_mean_{channel_str}p"].values.item()
        std_eta_star_an = d[f"eta_star_standard_deviation_{channel_str}a"].values.item()
        std_eta_star_pc = d[f"eta_star_standard_deviation_{channel_str}p"].values.item()
        fig_title_2 = rf"Analog calib. factor [{cal_height_an_min:3.1f} - {cal_height_an_max:3.1f} m] = {eta_star_an:3.2f} $\pm$ {std_eta_star_an:3.2f}"
        fig_title_3 = rf"Photocounting calib. factor [{cal_height_pc_min:3.1f} - {cal_height_pc_max:3.1f} m] = {eta_star_pc:3.2f} $\pm$ {std_eta_star_pc:3.2f}"

        plt.suptitle(f"{fig_title_1} \n {fig_title_2} \n {fig_title_3}", fontsize=16)

        # ANALOG
        signal_to_rcs(d[f"signal_T_P45_{channel_str}a"], d.range).plot(
            ax=ax1, y="range", lw=2, c="lime", label=r"RCS$^T_{+45}$. AN"
        )  # type: ignore
        signal_to_rcs(d[f"signal_R_N45_{channel_str}a"], d.range).plot(
            ax=ax1, y="range", lw=2, c="darkgreen", label=r"RCS$^R_{-45}$. AN"
        )  # type: ignore
        signal_to_rcs(d[f"signal_T_N45_{channel_str}a"], d.range).plot(
            ax=ax1, y="range", lw=2, c="r", label=r"RCS$^T_{-45}$. AN"
        )  # type: ignore
        signal_to_rcs(d[f"signal_R_P45_{channel_str}a"], d.range).plot(
            ax=ax1, y="range", lw=2, c="darkred", label=r"RCS$^R_{+45}$. AN"
        )  # type: ignore

        # PHOTONCOUNTING
        signal_to_rcs(d[f"signal_T_P45_{channel_str}p"], d.range).plot(ax=ax1, y="range", lw=2, c="deepskyblue", label=r"RCS$^T_{+45}$. PC")  # type: ignore
        signal_to_rcs(d[f"signal_R_N45_{channel_str}p"], d.range).plot(
            ax=ax1, y="range", lw=2, c="b", label=r"RCS$^R_{-45}$. PC"
        )  # type: ignore
        signal_to_rcs(d[f"signal_T_N45_{channel_str}p"], d.range).plot(
            ax=ax1, y="range", lw=2, c="magenta", label=r"RCS$^T_{-45}$. PC"
        )  # type: ignore
        signal_to_rcs(d[f"signal_R_P45_{channel_str}p"], d.range).plot(
            ax=ax1, y="range", lw=2, c="darkmagenta", label=r"RCS$^R_{+45}$. PC"
        )  # type: ignore

        ax1.grid()
        ax1.axes.set_xlabel(r"RCS [a.u.]")
        ax1.axes.set_ylabel(r"Height [km, agl]")
        ax1.axes.set_ylim(y_lim)  # type: ignore
        ax1.axes.set_xscale("log")
        ax1.legend(fontsize="small", loc="upper left")

        (d[f"signal_R_P45_{channel_str}a"] / d[f"signal_T_P45_{channel_str}a"]).plot(
            ax=ax2, y="range", lw=2, c="blue", label=r"AN at +45"
        )  # type: ignore
        (d[f"signal_R_N45_{channel_str}a"] / d[f"signal_T_N45_{channel_str}a"]).plot(
            ax=ax2, y="range", lw=2, c="red", label=r"AN at -45"
        )  # type: ignore
        (d[f"signal_R_P45_{channel_str}p"] / d[f"signal_T_P45_{channel_str}p"]).plot(
            ax=ax2, y="range", lw=2, c="green", label=r"PC at +45"
        )  # type: ignore
        (d[f"signal_R_N45_{channel_str}p"] / d[f"signal_T_N45_{channel_str}p"]).plot(
            ax=ax2, y="range", lw=2, c="black", label=r"PC at -45"
        )  # type: ignore

        ax2.grid()
        ax2.axes.set_xlabel(r"Signal Ratio [R / T]")
        ax2.axes.set_ylabel(r"")
        ax2.legend(fontsize="small", loc="upper left")
        ax2.axes.set_xlim(kwargs.get("signal_ratio_x_lim", (0, 10)))

        d[f"eta_star_profile_{channel_str}a"].plot(
            ax=ax3, y="range", c="g", label=r"AN"
        )  # type: ignore
        d[f"eta_star_profile_{channel_str}p"].plot(
            ax=ax3, y="range", c="m", label=r"PC"
        )  # type: ignore
        ax3.grid()
        ax3.axes.set_xlabel(r"Calibration Factor")
        ax3.axes.set_ylabel(r"")
        ax3.axes.set_xlim(kwargs.get("calib_factor_x_lim", (0, 5)))
        ax3.legend(fontsize="small", loc="upper left")
        # ax3.axes.set_xlim((0, 1))

        plt.tight_layout()
    except ValueError:
        output_file = None
        raise ValueError

    if output_dir.exists() and output_dir.is_dir():
        calibration_datetime = d.attrs["calibration_datetime"]
        output_file = (
            output_dir / f"calibration_{channel_str}_{calibration_datetime}.png"
        )   

        plt.savefig(output_file, dpi=300, bbox_inches="tight")
    else:
        output_file = None
        raise FileNotFoundError("`output_dir` not found.")
    return output_file



import os
import sys
import logging
import numpy as np
import xarray as xr
from pathlib import Path
from datetime import datetime
from abc import ABC

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
import matplotlib.colors as mcolors

from gfatpy.lidar.utils.utils import signal_to_rcs
from gfatpy.lidar.scc.plot import PLOT_INFO
from gfatpy.lidar.scc.plot.utils import id2str
from gfatpy.utils.plot import color_list


logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel(logging.INFO)


class SCC_product(ABC):
    """A simple class to manage generic SCC products."""

    def __init__(self, path: Path):
        self.path = self.check_path(path)

    def check_path(self, path: Path) -> Path:
        """Check if the path exists and is a netCDF file.

        Args:
            path (Path): Path to check.

        Raises:
            ValueError: path does not exist.
            ValueError: Filepath is not a netCDF file.

        Returns:
            Path: Path to the netCDF file.
        """
        if not path.exists():
            raise ValueError(f"{path} does not exist.")
        # check zip
        if path.suffix != ".nc":
            raise ValueError("Filepath is not a netCDF file.")
        return path

    def open_dataset(self) -> xr.Dataset:
        """Open the netCDF file.

        Returns:
            xr.Dataset: NetCDF file.
        """
        data = xr.open_dataset(self.path)
        return data


class SCC_raw(SCC_product):
    """A simple class to manage SCC raw products based on SCC_product."""

    def __init__(self, path: Path, range_resolution: float = 3.75):
        super().__init__(path)
        (
            self.station,
            self.datetime,
            self.measurementID,
            self.type,
        ) = self._get_info(self.path)
        self.range_resolution = range_resolution
        self.channels_ID = self._get_channel_ID()

    def _get_info(self, path: Path) -> tuple[str, datetime, str]:
        """Get information from SCC raw netcdf files.

        Args:
            path (Path): Path to the SCC raw netcdf file.

        Returns:
            tuple[str, datetime, int]: Information from the SCC raw netcdf file. In order: station, datetime, measurementID, scc_product_type.
        """
        # raw: raw_202307060159_202307060259_20230706gra0200_elpp_v5.2.7.nc
        measurementID = path.name.split("_")[-1].split(".")[0]
        ini_date, station, ini_hour = (
            measurementID[:8],
            measurementID[8:11],
            measurementID[11:],
        )
        ini_datetime = datetime.strptime(f"{ini_date}{ini_hour}", "%Y%m%d%H%M")
        return (
            station,
            ini_datetime,
            measurementID,
            "raw",
        )

    def _get_channel_ID(self) -> np.ndarray:
        """Get the channel ID from the netCDF file.

        Returns:
            np.ndarray: Channel ID.
        """
        with xr.open_dataset(self.path) as data:
            return data.channel_ID.values

    def channel2str(self, ID: int) -> str:
        """Provides the string of the channel/product ID.

        Args:
            ID (int): ID channel/product

        Returns:
            str: string describing the channel/product.
        """
        return id2str(ID)

    def _subtrack_dark_measurement(
        self, data: xr.Dataset, fill_value=9.969209968386869e36
    ) -> xr.Dataset:
        """It substract the dark measurment `Background_profile`.

        Args:
            data (xr.Dataset): data from self.data
            fill_value (_type_, optional): Fill value used in the netcdf. Defaults to EARLINET fill value: 9.969209968386869e+36.

        Returns:
            xr.Dataset: same Dataset but including `Raw_Lidar_Data_dc`.
        """
        if "Background_Profile" not in data:
            logger.warning(
                "Background_Profile not found in data. Dark measurement subtraction will not be performed."
            )
            return data
        channel_IDs = [
            channel_
            for channel_ in self.channels_ID
            if data["DAQ_Range"].sel(channels=channel_).values != fill_value
        ]
        data["Raw_Lidar_Data_dc"] = data["Raw_Lidar_Data"].copy()

        for channel_ in channel_IDs:
            # find the index of the channel channel_ID in data['channels']
            idx_ = np.where(data["channels"].values == channel_)[0][0]
            data["Raw_Lidar_Data_dc"][:, idx_, :] = data["Raw_Lidar_Data"].sel(
                channels=channel_
            ) - data["Background_Profile"].sel(channels=channel_).mean("time_bck")
        return data

    def _subtract_background(self, data: xr.Dataset) -> xr.Dataset:
        """It substract the background according to `Background_Mode`.

        Args:
            data (xr.Dataset): data from self.data.

        Returns:
            xr.Dataset: Same data including variable `Raw_Lidar_Data_bckg` or `Raw_Lidar_Data_dc__bckg`.
        """
        if "Raw_Lidar_Data_dc" not in data:
            raw_string = "Raw_Lidar_Data"
        else:
            raw_string = "Raw_Lidar_Data_dc"
        data[f"{raw_string}_bckg"] = data[raw_string].copy(
            data=np.nan * np.ones(data[raw_string].shape)
        )

        for channel_idx, channel_ in enumerate(self.channels_ID):
            if data["Background_Mode"].sel(channels=channel_).item() == 0:
                low_background = int(
                    data["Background_Low"].sel(channels=channel_).item()
                )
                high_background = int(
                    data["Background_High"].sel(channels=channel_).item()
                )
            else:
                low_background = int(
                    data["Background_Low"].sel(channels=channel_).item()
                    / self.range_resolution
                )
                high_background = int(
                    data["Background_High"].sel(channels=channel_).item()
                    / self.range_resolution
                )
            bckgrd_ = (
                data[raw_string]
                .sel(channels=channel_, points=slice(low_background, high_background))
                .mean("points")
            )
            data[f"{raw_string}_bckg"][:, channel_idx, :] = (
                data[raw_string].sel(channels=channel_) - bckgrd_
            )

        return data

    def _zerobin_correction(self, data: xr.Dataset) -> xr.Dataset:
        """It corrects from zero bin using `First_Signal_Rangebin`.

        Args:
            data (xr.Dataset): data from self.data

        Raises:
            ValueError: Background subtraction must be performed before zero bin correction.

        Returns:
            xr.Dataset: same data including `Raw_Lidar_Data_dc_bckg_zb` or `Raw_Lidar_Data_bckg_zb`.
        """
        raw_string = "Raw_Lidar_Data_dc_bckg"
        if raw_string not in data:
            raw_string = "Raw_Lidar_Data_bckg"
        if raw_string not in data:
            raise ValueError(
                "Background subtraction must be performed before zero bin correction."
            )

        shifted_data = np.nan * np.ones(data[raw_string].shape)
        topbin = data[raw_string].shape[-1]
        for channel_idx, channel_ in enumerate(data.channels.values):
            zerobin_ = data["First_Signal_Rangebin"].sel(channels=channel_).item()
            shifted_data[:, channel_idx, : (topbin - zerobin_)] = data[raw_string].sel(
                channels=channel_, points=slice(zerobin_, topbin)
            )

        # add shifted data to dataset as Raw_Lidar_Data_dc_bckg_zb
        data[f"{raw_string}_zb"] = (("time", "channels", "points"), shifted_data)
        return data

    def scc_preprocessing(self) -> xr.Dataset:
        """It preprocesses the raw data: 1) Dark measurement, 2) Background, 3) Zero bin. Background correction is mandatory.

        Returns:
            xr.Dataset: same data including 'Raw_Lidar_Data_dc_bckg_zb' or 'Raw_Lidar_Data_bckg_zb'
        """
        data = self.open_dataset()
        data["channels"] = data["channel_ID"].values
        data = self._subtrack_dark_measurement(data=data)
        data = self._subtract_background(data=data)
        data = self._zerobin_correction(data=data)
        return data

    def plot(
        self,
        rcs_limits: tuple[float, float] | None = None,
        range_limits: tuple[float, float] | None = None,
        output_dir: Path | None = None,
        dpi: int = 300,
        savefig=False,
        **kwargs,
    ) -> tuple[list[Figure], list[Path | None]]:

        if kwargs.get("preprocess", True):
            data = self.scc_preprocessing()
        else:
            data = self.open_dataset()
        data["points"] = (
            self.range_resolution * np.arange(1, data.dims["points"] + 1) / 1e3
        )
        figs, outputpaths = [], []
        for channel_ in data.channels.values:
            DAQ = data["DAQ_Range"].sel(channels=channel_).values.item()  # mV
            fill_value = 9.969209968386869e36
            fig, ax = plt.subplots(2, 1, figsize=(10, 5), sharex=True)
            if isinstance(ax, Axes):
                ax = [ax]
            string_ID = self.channel2str(channel_)
            wavelength = int(string_ID.split("_")[0][:-3])
            colors = color_list(
                data.dims["time"], cmap=PLOT_INFO["wavelength2cmaps"][wavelength]
            )
            raw_colors = color_list(data.dims["time"], cmap="Greys")
            for time_idx, time_ in enumerate(
                data["Raw_Data_Start_Time"].values.flatten()
            ):
                try:
                    if "Raw_Lidar_Data_dc_bckg_zb" in data:
                        signal_to_rcs(
                            data.sel(time=time_idx, channels=channel_)[
                                "Raw_Lidar_Data_dc_bckg_zb"
                            ],
                            data["points"] * 1e3,
                        ).plot(
                            x="points", ax=ax[0], color=colors[time_idx], linewidth=0.25
                        )
                        data.sel(time=time_idx, channels=channel_)[
                            "Raw_Lidar_Data_dc_bckg_zb"
                        ].plot(
                            x="points", ax=ax[1], color=colors[time_idx], linewidth=0.25
                        )
                        data.sel(time=time_idx, channels=channel_)[
                            "Raw_Lidar_Data"
                        ].plot(
                            x="points",
                            ax=ax[1],
                            color=raw_colors[time_idx],
                            linewidth=0.25,
                        )
                    elif "Raw_Lidar_Data_bckg_zb" in data:
                        signal_to_rcs(
                            data.sel(time=time_idx, channels=channel_)[
                                "Raw_Lidar_Data_bckg_zb"
                            ],
                            data["points"] * 1e3,
                        ).plot(
                            x="points", ax=ax[0], color=colors[time_idx], linewidth=0.25
                        )
                        data.sel(time=time_idx, channels=channel_)[
                            "Raw_Lidar_Data_bckg_zb"
                        ].plot(
                            x="points", ax=ax[1], color=colors[time_idx], linewidth=0.25
                        )
                        data.sel(time=time_idx, channels=channel_)[
                            "Raw_Lidar_Data"
                        ].plot(
                            x="points",
                            ax=ax[1],
                            color=raw_colors[time_idx],
                            linewidth=0.25,
                        )
                    else:
                        data.sel(time=time_idx, channels=channel_)[
                            "Raw_Lidar_Data"
                        ].plot(
                            x="points",
                            ax=ax[1],
                            color=raw_colors[time_idx],
                            linewidth=0.25,
                        )
                except:
                    raise ValueError(
                        f"Error plotting {self.path.name} | {string_ID} | {time_}"
                    )
            fig.suptitle(
                f"{self.station} | {string_ID} | {channel_} | {self.measurementID}"
            )
            ax[0].set_ylabel("Uncal. RCS, [a.u.]")
            if rcs_limits is not None:
                ax[0].set_ylim(rcs_limits)
            if data["DAQ_Range"].sel(channels=channel_).values != fill_value:
                if kwargs.get("analog_limits", None) is not None:
                    ax[1].set_ylim(kwargs.get("analog_limits"))
                ax[1].set_ylabel("Signal, [mV]")
                ax[1].axhline(y=DAQ / 20000, color="r", linestyle="--", linewidth=1)
            else:
                if kwargs.get("photoncounting_limits", None) is not None:
                    ax[1].set_ylim(kwargs.get("photoncounting_limits"))
                ax[1].axhline(y=20, color="b", linestyle="--", linewidth=1)
                ax[1].set_ylabel("Signal, [MHz]")
            if range_limits is not None:
                ax[0].set_xlim(range_limits)
            else:
                ax[0].set_xlim(0, 15)
            # Remove xlabel from ax[0]
            ax[0].set_xlabel("")
            ax[1].set_xlabel("Range, [km]")
            ax[0].set_yscale("log")
            for ax_ in ax:
                ax_.minorticks_on()
                ax_.set_title("")
                ax_.grid()
                # ax_.legend().remove()
            try:
                fig.tight_layout()
            except Exception as e:
                print(e)
                logger.warning(
                    f"Error in tight_layout for {self.path.name} | {string_ID}"
                )
                continue
            if savefig:
                # save
                if isinstance(output_dir, Path):
                    output_dir.mkdir(parents=True, exist_ok=True)
                elif output_dir is None:
                    output_dir = Path.cwd()
                outputpath = output_dir.joinpath(
                    self.path.name.replace(".nc", f"_{string_ID}.png")
                )
                fig.savefig(outputpath, dpi=dpi)
                data.close()
                outputpaths.append(outputpath)
            figs.append(fig)
        data.close()
        return figs, outputpaths


class SCC_elpp(SCC_product):
    """A simple class to manage SCC elpp products based on SCC_product."""

    def __init__(self, path: Path):
        super().__init__(path)
        (
            self.station,
            self.number1,
            self.productID,
            self.datetime_ini,
            self.datetime_end,
            self.measurementID,
            self.type,
            self.scc_version,
            self.hoi_configuration_ID,
        ) = self._get_info(self.path)
        self.string_ID = id2str(self.productID)

    def _get_info(
        self, path: Path
    ) -> tuple[str, int, int, datetime, datetime, str, str, str, int]:
        """Get information from SCC ELPP netcdf files.

        Args:
            path (Path): Path to the SCC ELPP netcdf file.

        Returns:
            tuple[str, int, int, datetime, datetime, str, str, str, int]: Information from the SCC ELPP netcdf file. In order: station, number1, productID, ini_datetime, end_datetime, measurementID, scc_product_type, scc_version_str, hoi_configuration_ID.
        """
        # elpp: gra_003_0002223_202307060159_202307060259_20230706gra0200_elpp_v5.2.7.nc
        (
            station,
            number1,
            productID,
            ini_datestr,
            end_datestr,
            measurementID,
            scc_product_type,
            scc_version_str,
        ) = path.name.split("_")
        ini_datetime = datetime.strptime(ini_datestr, "%Y%m%d%H%M")
        end_datetime = datetime.strptime(end_datestr, "%Y%m%d%H%M")
        with xr.open_dataset(self.path) as ds:
            hoi_configuration_ID = ds.attrs["hoi_configuration_ID"]

        return (
            station,
            int(number1),
            int(productID),
            ini_datetime,
            end_datetime,
            measurementID,
            scc_product_type,
            scc_version_str[:-3],
            hoi_configuration_ID,
        )

    def rayleight_fit(
        self,
        normalization_range: tuple[float, float] = (6, 7),
        range_limits: tuple[float, float] = (0, 25),
        attenuated_backscatter_limits: tuple[float, float] | None = None,
        output_dir: Path | None = None,
        dpi: int = 300,
        combine_channels: bool = False,
        savefig: bool = False,
    ) -> tuple[Figure, Path | None]:
        """plot Rayleigh fit from SCC ELPP netcdf files.

        Args:
            normalization_range (tuple[float, float], optional): Minimum and maximum ranges to normalized signals. Defaults to (6, 7).
            range_limits (tuple[float, float], optional): Range limits. Defaults means (0, 25).
            attenuated_backscatter_limits (tuple[float, float] | None, optional): Attenuated backscatter limits. Defaults to None means automatic limits from `gfatpy.lidar.scc.plot.PLOT_INFO`.
            savefig (bool, optional): If True save figure. Defaults to False.
            output_dir (Path | None, optional): Output directory to save figure. Defaults to None means current working directory.
            dpi (int, optional): _description_. Defaults to 300.
            combine_channels (bool, optional): Combine in the same figure all the signals linked to a givel ELPP product. Defaults to False.

        Returns:
            list[Figure]: List of figures.
        """

        station, hoi_configuration_ID, string_ID = (
            self.station,
            self.hoi_configuration_ID,
            self.string_ID,
        )
        raw_data = self.open_dataset()
        raw_data["altitude"] = raw_data["altitude"] / 1e3
        attenuated_molecular_backscatter = (
            (raw_data.molecular_extinction / raw_data.molecular_lidar_ratio)
            * raw_data.molecular_transmissivity_at_detection_wavelength
            * raw_data.molecular_transmissivity_at_emission_wavelength
        )
        if combine_channels:
            fig, ax = plt.subplots(figsize=(10, 4))
        for idx, channel_ in enumerate(raw_data.channel):
            if not combine_channels:
                fig, ax = plt.subplots(figsize=(8, 4))
            IDs = raw_data.range_corrected_signal_channel_id.sel(
                channel=channel_
            ).values.flatten()
            string_list = [id2str(int(ID)) for ID in IDs]
            channelID_ = os.path.commonprefix(string_list)
            wavelength = int(string_list[0].split("_")[0][:-3])
            if len(raw_data.time) > 1:
                colors = color_list(
                    raw_data.dims["time"],
                    cmap=PLOT_INFO["wavelength2cmaps"][wavelength],
                )
            else:
                if "b" in string_ID:
                    colors = [
                        tuple(
                            np.asarray(
                                list(
                                    mcolors.to_rgba(
                                        PLOT_INFO["backscatter2color"][self.string_ID]
                                    )
                                )
                            )[:3]
                            / (idx + 1)
                        )
                        + tuple([1])
                    ]
                elif "lre" in string_ID or "eo" in string_ID:
                    colors = [
                        tuple(
                            np.asarray(
                                list(
                                    mcolors.to_rgba(
                                        PLOT_INFO["extinction2color"][self.string_ID]
                                    )
                                )
                            )[:3]
                            / (idx + 1)
                        )
                        + tuple([1])
                    ]
                else:
                    colors = ["orange"]
            for idx, time_ in enumerate(raw_data.time.values):
                range_resol = (
                    raw_data.altitude.sel(time=time_)[1]
                    - raw_data.altitude.sel(time=time_)[0]
                )
                normalization_index = [
                    int(i / range_resol) for i in normalization_range
                ]
                norm_attbetamol = attenuated_molecular_backscatter.sel(
                    level=slice(*normalization_index)
                ).mean("level")
                norm_rcs = (
                    raw_data["range_corrected_signal"]
                    .sel(level=slice(*normalization_index))
                    .mean("level")
                )
                normalized_rcs = raw_data["range_corrected_signal"] * (
                    norm_attbetamol / norm_rcs
                )
                (1e6 * normalized_rcs.sel(time=time_, channel=channel_)).plot(
                    x="altitude",
                    ax=ax,
                    label=r"$\beta_{RCS}^{att}$"
                    + f"[{channelID_}] | {time_.astype(str).split('.')[0]}",
                    color=colors[idx],
                    linewidth=1.0,
                )
                (
                    1e6
                    * attenuated_molecular_backscatter.sel(time=time_, channel=channel_)
                ).plot(
                    x="altitude",
                    ax=ax,
                    label=r"$\beta_{mol}^{att}$"
                    + f"[{wavelength} nm] | {time_.astype(str).split('.')[0]}",
                    color="orange",
                    linewidth=1.0,
                )
            ax.set_xlabel("Range, [km]")
            ax.set_ylabel("Att. backscatter, [Mm$^{-1}$sr$^{-1}$]")
            ax.set_title(
                f"{station} | Lidar ID: {hoi_configuration_ID} | Product ID/name: {self.productID}/{string_ID} \n Norm. range: {normalization_range} km"
            )
            ax.set_yscale("log")
            ax.set_xlim(*range_limits)
            if attenuated_backscatter_limits is not None:
                ax.set_ylim(*attenuated_backscatter_limits)
            else:
                ax.set_ylim(*PLOT_INFO["limits"]["attenuated_backscatter"][wavelength])
            ax.legend(fontsize=10, loc="upper right")
            ax.minorticks_on()
            ax.grid()
            ax.set_facecolor("white")
            fig.tight_layout()
            if savefig:
                if output_dir is None:
                    output_dir = Path.cwd()
                output_dir = Path(output_dir)
                output_dir.mkdir(exist_ok=True)
                outputpath = output_dir / self.path.name.replace("nc", "png")
                fig.savefig(
                    outputpath,
                    dpi=dpi,
                    bbox_inches="tight",
                )
            else:
                outputpath = None
        raw_data.close()
        return fig, outputpath

    def plot(
        self,
        rcs_limits: tuple[float, float] | None = None,
        range_limits: tuple[float, float] | None = None,
        output_dir: Path | None = None,
        dpi: int = 300,
        savefig=False,
    ) -> tuple[Figure, Path | None]:
        """Plot SCC ELPP netcdf files.

        Args:
            rcs_limits (tuple[float, float] | None, optional): Range Corrected Signal limits. Defaults to None. None means automatic limits.
            range_limits (tuple[float, float] | None, optional): Range limits. Defaults to None. None means automatic limits.
            output_dir (Path | None, optional): Output directory. Defaults to None. None means current working directory.
            dpi (int, optional): Plot resolution. Defaults to 300.

        Raises:
            ValueError: Filepath does not exist.
            ValueError: Error plotting.

        Returns:
            Path: Path to the plot.
        """

        if not Path(self.path).exists():
            raise ValueError("Filepath does not exist.")

        # Load netcdf
        elpp = xr.open_dataset(self.path)
        elpp["altitude"] = elpp["altitude"] / 1e3
        hoi_configuration_ID = elpp.attrs["hoi_configuration_ID"]
        fig, ax = plt.subplots(1, 1, figsize=(10, 5))

        for product_ in elpp.channel.values:
            IDs = elpp.range_corrected_signal_channel_id.sel(
                channel=product_
            ).values.flatten()
            string_list = [id2str(int(ID)) for ID in IDs]
            channelID_ = os.path.commonprefix(string_list)
            wavelength = int(string_list[0].split("_")[0][:-3])
            colors = color_list(
                elpp.dims["time"], cmap=PLOT_INFO["wavelength2cmaps"][wavelength]
            )
            for idx, time_ in enumerate(elpp.time.values):
                try:
                    elpp.sel(time=time_, channel=product_)[
                        "range_corrected_signal"
                    ].plot(
                        x="altitude",
                        ax=ax,
                        label=f"{channelID_} | {time_.astype(str).split('.')[0]}",
                        color=colors[idx],
                        linewidth=1.0,
                    )
                except:
                    raise ValueError(
                        f"Error plotting {self.path.name} | {product_} | {time_}"
                    )
        ax.set_title(f"{self.station} | {hoi_configuration_ID} | {self.string_ID}")
        if rcs_limits is not None:
            ax.set_ylim(rcs_limits)
        if range_limits is not None:
            ax.set_xlim(range_limits)
        else:
            ax.set_xlim(0, 15)
        ax.set_yscale("log")
        ax.legend(fontsize=10, loc="upper right")
        ax.minorticks_on()
        ax.grid()
        ax.set_xlabel("Range, [km]")
        ax.set_ylabel("Uncal. RCS, [a.u.]")
        fig.tight_layout()
        if savefig:
            # save
            if isinstance(output_dir, Path):
                output_dir.mkdir(parents=True, exist_ok=True)
            elif output_dir is None:
                output_dir = Path.cwd()
            outputpath = output_dir.joinpath(self.path.name.replace(".nc", ".png"))
            fig.savefig(outputpath, dpi=dpi)
            elpp.close()
            plt.close(fig)
        else:
            outputpath = None
        return fig, outputpath


class SCC_elda(SCC_product):
    """A simple class to manage SCC elda products based on SCC_product."""

    def __init__(self, path: Path):
        super().__init__(path)
        (
            self.station,
            self.number1,
            self.wavelength,
            self.productID,
            self.datetime_ini,
            self.datetime_end,
            self.measurementID,
            self.type,
            self.scc_version,
            self.hoi_configuration_ID,
        ) = self._get_info(self.path)
        self.string_ID = id2str(self.productID)
        self.earlinet_product_type = self._earlinet_product_type()

    def _get_info(
        self, path: Path
    ) -> tuple[str, int, int, int, datetime, datetime, str, str, str, int]:
        """Get information from SCC ELDA netcdf files.

        Args:
            path (Path): Path to the SCC ELDA netcdf file.

        Returns:
            tuple[str, int, int, int, datetime, datetime, str, str, str, int]: Information from the SCC ELDA netcdf file. In order: station, number1, wavelength, productID, ini_datetime, end_datetime, measurementID, scc_product_type, scc_version_str, hoi_configuration_ID.
        """
        # elda: gra_001_0532_0002366_202302220029_202302220059_20230222gra0000_elda_v5.2.7
        (
            station,
            number1,
            wavelength,
            productID,
            ini_datestr,
            end_datestr,
            measurementID,
            scc_product_type,
            scc_version_str,
        ) = path.name.split("_")
        ini_datetime = datetime.strptime(ini_datestr, "%Y%m%d%H%M")
        end_datetime = datetime.strptime(end_datestr, "%Y%m%d%H%M")
        with xr.open_dataset(self.path) as ds:
            hoi_configuration_ID = ds.attrs["hoi_configuration_ID"]

        return (
            station,
            int(number1),
            int(wavelength),
            int(productID),
            ini_datetime,
            end_datetime,
            measurementID,
            scc_product_type,
            scc_version_str[:-3],
            hoi_configuration_ID,
        )

    def _earlinet_product_type(self) -> dict:
        """Return the EARLINET product flags.

        Args:
            elda_product (SCC_elda): SCC elda object.

        Returns:
            dict: Dictionary with numbers [int] as keys and its meaning [string] as value.
        """
        with xr.open_dataset(self.path) as elda:
            flag_dict = dict(
                zip(
                    elda.earlinet_product_type.flag_values,
                    elda.earlinet_product_type.flag_meanings.split(" "),
                )
            )

        return flag_dict[elda.earlinet_product_type.values.item()]

    def plot_backscatter(
        self,
        ax: Axes,
        backscatter_limits: tuple[float, float] | None = None,
        range_limits: tuple[float, float] | None = None,
    ) -> Axes:
        """Plot backscatter from SCC elda netcdf files.

        Args:
            ax (Axes): Axes to plot.
            backscatter_limits (tuple[float, float] | None, optional): Backcatter coefficient limits. Defaults to None. None means automatic limits.
            range_limits (tuple[float, float] | None, optional): Range limits. Defaults to None. None means automatic limits.

        Raises:
            ValueError: Error plotting.

        Returns:
            Axes: Axes with the plot.
        """

        # Load netcdf
        elda = xr.open_dataset(self.path)
        elda["altitude"] = elda["altitude"] / 1e3
        for time_ in elda.time.values:
            try:
                (elda.sel(time=time_)["backscatter"] * 1e6).plot(
                    y="altitude",
                    ax=ax,
                    label=self.string_ID,
                    color=PLOT_INFO["backscatter2color"][self.string_ID],
                    linewidth=1.0,
                )
            except:
                raise ValueError(f"Error plotting {self.path.name} | {time_}")
        ax.axvline(x=0, color="k", linestyle="--", linewidth=1)
        if backscatter_limits is not None:
            ax.set_xlim(*backscatter_limits)
        if range_limits is not None:
            ax.set_xlim(*PLOT_INFO["limits"]["backscatter"])
        ax.set_xlabel(r"$\beta_{part}$, [$Mm^{-1}$]")
        elda.close()
        return ax

    def plot_extinction(
        self,
        ax: Axes,
        extinction_limits: tuple[float, float] | None = None,
        range_limits: tuple[float, float] | None = None,
    ) -> Axes:
        """Plot extinction from SCC elda netcdf files.

        Args:
            ax (Axes): Axes to plot.
            extinction_limits (tuple[float, float] | None, optional): Extinction limits. Defaults to None. None means automatic limits.
            range_limits (tuple[float, float] | None, optional): Range limits. Defaults to None. None means automatic limits.

        Raises:
            ValueError: Error plotting.

        Returns:
            Axes: Axes with the plot.
        """

        # Load netcdf
        elda = xr.open_dataset(self.path)
        elda["altitude"] = elda["altitude"] / 1e3
        color = PLOT_INFO["extinction2color"][self.string_ID]
        for time_ in elda.time.values:
            try:
                (elda.sel(time=time_)["extinction"] * 1e6).plot(
                    y="altitude",
                    ax=ax,
                    label=self.string_ID,
                    color=color,
                    linewidth=1.0,
                )
            except:
                raise ValueError(f"Error plotting {self.path.name} | {time_}")
        # plot vertical line at zero
        ax.axvline(x=0, color="k", linestyle="--", linewidth=1)
        if extinction_limits is not None:
            ax.set_xlim(*extinction_limits)
        else:
            ax.set_ylim(*PLOT_INFO["limits"]["extinction"])
        if range_limits is not None:
            ax.set_ylim(range_limits)
        ax.set_xlabel(r"$\alpha_{part}$, [$Mm^{-1}$]")
        elda.close()
        return ax

    def plot_depo(
        self,
        ax: Axes,
        depo_limits: tuple[float, float] | None = None,
        range_limits: tuple[float, float] | None = None,
    ) -> Axes:
        """Plot depolarization from SCC elda netcdf files.

        Args:
            ax (Axes): Axes to plot.
            depo_limits (tuple[float, float] | None, optional): Depolarization limits. Defaults to None. . None means limits `from gfatpy.lidar.plot.info.yml`.
            range_limits (tuple[float, float] | None, optional): Range limits. Defaults to None. None means automatic limits.

        Raises:
            ValueError: Error plotting.

        Returns:
            Axes: Axes with the plot.
        """

        # Load netcdf
        elda = xr.open_dataset(self.path)
        elda["altitude"] = elda["altitude"] / 1e3
        color = PLOT_INFO["depolarization2color"][self.string_ID]
        for time_ in elda.time.values:
            try:
                elda.sel(time=time_)["particledepolarization"].plot(
                    y="altitude",
                    ax=ax,
                    label=self.string_ID,
                    color=color,
                    linewidth=2.5,
                )
                elda.sel(time=time_)["volumedepolarization"].plot(
                    y="altitude",
                    ax=ax,
                    label=self.string_ID,
                    color=color,
                    linewidth=1.0,
                    linestyle="--",
                )
            except:
                raise ValueError(f"Error plotting {self.path.name} | {time_}")
        # plot vertical line at zero
        ax.axvline(x=0, color="k", linestyle="--", linewidth=1)
        if depo_limits is not None:
            ax.set_xlim(*depo_limits)
        else:
            ax.set_xlim(*PLOT_INFO["limits"]["depolarization"])
        if range_limits is not None:
            ax.set_ylim(range_limits)
        ax.set_xlabel(r"$\delta_{part}$, [#]")
        elda.close()
        return ax

    def plot_lidar_ratio(
        self,
        ax: Axes,
        lidar_ratio_limits: tuple[float, float] | None = None,
        range_limits: tuple[float, float] | None = None,
    ) -> Axes:
        """Plot lidar ratio from SCC elda netcdf files.

        Args:
            ax (Axes): Axes to plot.
            lidar_ratio_limits (tuple[float, float] | None, optional): Lidar ratio limits. Defaults to None. None means limits `from gfatpy.lidar.plot.info.yml`.
            range_limits (tuple[float, float] | None, optional): Range limits. Defaults to None. None means automatic limits.

        Raises:
            ValueError: Error plotting.

        Returns:
            Axes: Axes with the plot.
        """

        # Load netcdf
        elda = xr.open_dataset(self.path)
        elda["altitude"] = elda["altitude"] / 1e3
        color = PLOT_INFO["lidar_ratio2color"][self.string_ID]
        for time_ in elda.time.values:
            try:
                (
                    elda.sel(time=time_)["extinction"]
                    / elda.sel(time=time_)["backscatter"]
                ).plot(
                    y="altitude",
                    ax=ax,
                    label=self.string_ID,
                    color=color,
                    linewidth=1.0,
                )
            except:
                raise ValueError(f"Error plotting {self.path.name} | {time_}")
        # plot vertical line at zero
        ax.axvline(x=0, color="k", linestyle="--", linewidth=1)
        if lidar_ratio_limits is not None:
            ax.set_xlim(*lidar_ratio_limits)
        else:
            ax.set_xlim(*PLOT_INFO["limits"]["lidar_ratio"])
        if range_limits is not None:
            ax.set_ylim(range_limits)
        ax.set_xlabel(r"$LR_{part}$, [sr]")
        elda.close()
        return ax



#!/usr/bin/env python
import os
import glob
import pathlib
from pdb import set_trace
import warnings

import datetime as dt
import numpy as np
import xarray as xr
from loguru import logger

from gfatpy import utils
from gfatpy.lidar.utils.file_manager import filename2info
from gfatpy.utils.io import read_yaml_from_info
from gfatpy.lidar.utils.utils import LIDAR_INFO

warnings.filterwarnings("ignore")


__author__ = "Bravo-Aranda, Juan Antonio"
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Bravo-Aranda, Juan Antonio"
__email__ = "jabravo@ugr.es"
__status__ = "Production"

""" DEFAULT AUXILIAR INFO
"""
# Root Directory (in NASGFAT)  according to operative system


def reader_xarray(
    filelist: list[str] | str | pathlib.Path,
    date_ini: str | None = None,
    date_end: str | None = None,
    ini_range: float | None = None,
    end_range: float | None = None,
    percentage_required: int = 80,
    channels: list[str] = [],
) -> xr.Dataset:
    """
    Lidar data reader using xarray module.
    Inputs:
    - filelist: String with pattern for create a List of lidar files (i.e, '/drives/c/*.nc') (str)
                List of lidar files (list)
                Pathlib.Path object (pathlib.Path)
    - date_ini: 'yyyymmddThhmmss'
    - date_end: 'yyyymmddThhmmss'
    - ini_range: int/float (m)
    - end_range: int/float (m)
    - percentage_required= percentage of the time period required to continue the process. Default 80%  (int)
    - channels: list of channel number (e.g., [0, 1, 5]) or [] to load all of them
    Output:
    - lidar: dictionary or 'None' in case of error.
    """

    """ Aux Functions
    """

    #Global Variables within this function
    global INFO
    
    def add_required_channels(
        lidar_name: str,
        channels: list | np.ndarray,
    ) -> list:

        if isinstance(channels, np.ndarray):
            channels = channels.tolist()

        required_channels = channels.copy()
        for channel_ in channels:
            if channel_ in INFO["product_channels"].keys():
                required_channels = required_channels + [
                    *INFO["product_channels"][channel_]
                ]
        unique_channels = np.unique(np.array(required_channels)).tolist()
        return unique_channels

    def select_channels(dataset: xr.Dataset, channels: list | str) -> xr.Dataset:
        """select_channels function creates a new dataset with 'signal_CHANNEL' defined in 'channels' (list).

        Args:
            dataset (xr.Dataset): lidar dataset
            channels (list | str): list of lidar channel names

        Returns:
            xr.Dataset: lidar dataset
        """

        if len(channels) > 0:
            if isinstance(channels, str):
                channels = [channels]
            if isinstance(channels, np.ndarray):
                channels = channels.tolist()

            # find variables: signal related to channel
            _vars = ["signal"]
            real_channels = []
            for _channel in dataset["channel"].values.tolist():
                if _channel not in channels:
                    for _var in _vars:
                        varname = "%s_%s" % (_var, _channel)
                        dataset = dataset.drop_vars(varname)
                else:
                    real_channels.append(_channel)
            dataset = dataset.sel(channel=real_channels)
            dataset = dataset.assign_coords(channel=real_channels)
        return dataset

    def check_minimum_profiles(
        times: np.ndarray,
        date_ini: dt.datetime,
        date_end: dt.datetime,
        percentage_required: float,
    ) -> bool:
        """Check Lidar Data has enough profiles

        Args:
            times ([type]): [description]
            date_ini ([type]): [description]
            date_end ([type]): [description]
            percentage_required ([type]): [description]
        """

        check = True
        time_resolution = float(
            np.median(np.diff(times)) / np.timedelta64(1, "s")
        )  # FIXME: typing error
        interval_duration = (date_end - date_ini).total_seconds()
        Nt = np.round(
            interval_duration / time_resolution
        )  # Theoretical Number of profiles
        Nm = (percentage_required / 100) * Nt  # Minimum number of profiles
        Np = len(times)  # Number of actual profiles
        if Np > Nm:
            logger.info(
                f"Data loaded from {date_ini.isoformat()} to {date_end.isoformat()}"
            )
        else:
            logger.warning(
                f"Not enough data found ({Np}<{Nm}) in the user-required period ({interval_duration} s.)"
            )
            check = False

        return check

    """ The Reader
        The Reader does:
        1. concatenate along time dimension
        2. merge channels comming from different telescopes (ALHAMBRA), assuming same range coordinate
    """
    logger.info("Start Reader ...")

    # Find Files to Read    
    try:
        if isinstance(filelist, str):            
            i_ = pathlib.Path(filelist)
            if i_.is_file():
                files2load = [str(i_.absolute())]
            else:
                file_list_ = [*i_.parent.rglob(i_.name)]
                files2load = [str(i_.absolute()) for i_ in file_list_]                
        elif isinstance(filelist, pathlib.Path):
            files2load = [str(filelist.absolute())]
        elif isinstance(filelist, list):        
            files2load = filelist
        else:            
            raise ValueError("filelist must be a string, pathlib.Path or list of strings.")
    except Exception as e:        
        files2load = []
        logger.warning(str(e))
        logger.warning(f"Files in {filelist} not found.")
    
    if len(files2load) > 0:
        logger.info(files2load)
        lidartemp = None        

        # Get lidar info from filename
        lidar_nick, _, _, _, _, date  = filename2info(pathlib.Path(files2load[0]).name)
        
        INFO = read_yaml_from_info(lidar_nick, date)


        if lidar_nick is None:
            logger.critical("Lidar nick not in lidar systems availables.")

        try:
            # Get lidar name
            lidar_name = LIDAR_INFO["metadata"]["nick2name"][lidar_nick]

            # Add required channels to the channel list to obtain product channel:
            channels = add_required_channels(lidar_name, channels)

            # Loop over modules: 1) concat time; 2) merge module
            lidar_ = None
            for fn in files2load:
                with xr.open_dataset(
                    fn, chunks={}
                ) as _dx:  # chunks={"time": 600, "range": 1000})
                    _dx = select_channels(_dx, channels)
                if not lidar_:
                    lidar_ = _dx
                else:
                    # concat only variables that have "time" dimension.
                    # rest of variables keep values from first dataset
                    try:
                        lidar_ = xr.concat(
                            [lidar_, _dx],
                            dim="time",
                            data_vars="minimal",
                            coords="minimal",
                            compat="override",
                        )
                    except Exception as e:
                        logger.critical("Dataset in {fn} not concatenated")
                        raise e
            # Sort Dataset by Time
            if lidar_ is not None:
                lidar_ = lidar_.sortby(lidar_["time"])
            else:
                raise ValueError("lidarmod is None.")
            # Merge Module
            if not lidartemp:
                lidartemp = lidar_
            else:
                try:
                    lidartemp = xr.merge([lidartemp, lidar_])
                except Exception as e:
                    logger.critical(f"{e}")
                    logger.critical(f"Datasets not merged")
                    raise e
        except Exception as e:
            logger.critical(f"{e}")
            logger.critical("Files not concatenated")
            raise e

        if lidartemp:
            # Selection time window and Check Enough Profiles
            if np.logical_and(date_ini is not None, date_end is not None):
                if np.logical_and(isinstance(date_ini, str), isinstance(date_end, str)):
                    # Times Formatting

                    date_ini_dt = utils.utils.str_to_datetime(date_ini)  # type: ignore
                    date_end_dt = utils.utils.str_to_datetime(date_end)  # type: ignore

                    # Time Selection
                    min_time_resol = dt.timedelta(seconds=0.1)
                    lidar = lidartemp.sel(
                        time=slice(
                            date_ini_dt - min_time_resol, date_end_dt + min_time_resol
                        )
                    )

                    # Check selection
                    ok = check_minimum_profiles(
                        lidar["time"].values,
                        date_ini_dt,
                        date_end_dt,
                        percentage_required,
                    )
                    if not ok:
                        lidar = None
                else:
                    lidar = lidartemp
            else:
                lidar = lidartemp
            del lidartemp

            # Complete lidar dataset
            if lidar:
                # Range Clip
                if ini_range is not None and end_range is not None:
                    if end_range > ini_range:
                        lidar = lidar.sel(range=slice(ini_range, end_range))
                    else:
                        raise ValueError("ini_range is larger than end_range.")

                # add background ranges
                if "BCK_MIN_ALT" not in lidar.attrs.keys():
                    lidar.attrs["BCK_MIN_ALT"] = 75000
                if "BCK_MAX_ALT" not in lidar.attrs.keys():
                    lidar.attrs["BCK_MAX_ALT"] = 105000

                # Extract information from filename
                try:
                    lidar.attrs["lidarNick"] = os.path.basename(files2load[0]).split(
                        "_"
                    )[0]
                    lidar.attrs["dataversion"] = os.path.basename(files2load[0]).split(
                        "_"
                    )[1]
                except:
                    lidar.attrs["lidarNick"] = "Unknown"
                    lidar.attrs["dataversion"] = "Unknown"

            else:
                lidar = None
        else:
            logger.error("Impossible to load found files.")
            lidar = None
    else:
        logger.error("Files not found.")
        lidar = None

    if lidar is None:
        # set_trace()
        logger.error("No dataset created.")
        raise RuntimeError("No dataset created.")

    logger.info("End Reader")
    return lidar





from typing import Tuple
from gfatpy.lidar.scc import scc_access

def check_scc_connection(scc_server_settings) -> bool: 
    """ Check if SCC server is available. """

    scc_obj = scc_access.SCC(
        tuple(scc_server_settings["basic_credentials"]),
        None,
        scc_server_settings["base_url"],
    )

    connection = scc_obj.login(scc_server_settings["website_credentials"])
    check_connection = connection.status_code == 200
    if not check_connection:
        raise Exception("Connection to SCC server failed.")
    else:
        scc_obj.logout()

    return check_connection

def check_measurement_id_in_scc(scc_server_settings, measurement: str) -> Tuple[bool, scc_access.Measurement | None]:
    """ Check if a measurement is already in SCC. """

    if len(measurement) != 15:
        raise ValueError("Measurement ID must have 14 characters.")

    scc_obj = scc_access.SCC( tuple(scc_server_settings["basic_credentials"]), None, scc_server_settings["base_url"], )

    scc_obj.login(scc_server_settings["website_credentials"])
    
    try:
        meas_obj, _ = scc_obj.get_measurement(measurement)
        measurement_id_in_scc = meas_obj is not None        
    except Exception as e:
        meas_obj = None
        number = int(str(e).split("Status code ")[-1][:-1])
        scc_obj.logout()
        raise ValueError(f"Error code: {number}")
    
    scc_obj.logout()

    return measurement_id_in_scc, meas_obj



from pathlib import Path

from . import rpg_nc, utils, types
from gfatpy.utils.io import read_yaml
from .plot import RADAR_PLOT_INFO
from .types import RadarInfoType


RADAR_INFO: RadarInfoType = read_yaml(Path(__file__).parent.absolute() / 'info.yml')

__all__ = ["RADAR_PLOT_INFO", "RADAR_INFO", "rpg_nc", "utils", "plot", "types"]


__doc__ = """The top of the radar module that is compatible (at least) with GFAT radar: "NEPHELE" AND "NEBULA"

    Contributors:
    
    -   Dr. Leonie	von Terzi
"""



import os
import shutil
from loguru import logger


def move2odir(source_filename, destination, overwrite=True):
    """
    It moves the file named source_filename in local directory to the absolute path file 'moved_filepath'.
    Input:
    source_filename: full path filename (string). 
    destination: destination path (string). 
    Output:
    sent: control variable. True=sent; False=not sent. (boolean).
    """
    sent = False    
    logger.info('Moving %s to %s' % (source_filename, destination))
    try:
        # create directory if it does not exist
        if not os.path.exists(destination):
            os.makedirs(destination)

        # delete file if exists in destination
        if os.path.exists(os.path.join(destination, os.path.basename(source_filename))):
            os.remove(os.path.join(destination, os.path.basename(source_filename)))

        # copyfile (shutil da problemas de incoherencia entre ejecucion (bien) y mensaje (error))
        sent = shutil.copy(source_filename, destination)
        if sent:
            sent = True
        logger.info('Moving file %s to user-defined output directory: %s... DONE!' % (source_filename, destination))

        # delete original
        os.remove(source_filename)
    except:
        if os.path.exists(os.path.join(destination, os.path.basename(source_filename))):
            logger.info('Moving file %s to user-defined output directory: %s... DONE!' % (source_filename, destination))
            sent = True
            # delete original
            os.remove(source_filename)
        else:
            logger.warning('Moving file %s to user-defined output directory: %s... ERROR!' % (source_filename, destination))
    return sent


""" Code to read Raymetrics version of Licel binary files."""
import logging

import matplotlib as mpl
import numpy as np
from matplotlib import pyplot as plt

from .licel import LicelFile, LicelLidarMeasurement, LicelChannel, PhotodiodeChannel

logger = logging.getLogger(__name__)


class ScanningFile(LicelFile):
    """ Raymetrics is using a custom version of licel file format to store scanning lidar measurements.
     
    The file includes one extra line describing the scan strategy of the dataset. The extra parameters are:
    
    `azimuth_start`
        Start azimuth angle for the scan, relative to instrument zero position (degrees).
    
    `azimuth_stop`
        Stop azimuth angle for the scan, relative to instrument zero position (degrees).
        
    `azimuth_step` 
        Step of the azimuth scan (degrees).

    `zenith_start` 
        Start zenith angle for the scan, relative to *nadir* (degrees). Take care that this is actually
        nadir angle. Vertical measurements correspond to -90.

    `zenith_stop` 
        Stop zenith angle for the scan, relative to *nadir* (degrees). Take care that this is actually
        nadir angle. Vertical measurements correspond to -90.

    `zenith_step` 
        Step of the zenith scan (degrees).

    `azimuth_offset`
        Offset of instrument zero from North (degrees). Using this value you can convert `azimuth_start` and
        `azimuth_stop` to absolute values.

    Moreover, four new parameters are added in the second line of the file:

    `zenith_angle`
        Zenith angle of the current file. Take care that this is actually
        nadir angle. Vertical measurements correspond to -90.

    `azimuth_angle`
        Azimuth angle of the current file. Value relative to instrument zero position.

    `temperature`
        Ambient temperature (degrees C)

    `pressure`
        Ambient pressure (hPa)
    """

    # Specifications of the header lines.
    licel_file_header_format = ['filename',
                                'start_date start_time end_date end_time altitude longitude latitude zenith_angle azimuth_angle temperature pressure custom_info',
                                # Appart from Site that is read manually
                                'azimuth_start azimuth_stop azimuth_step zenith_start zenith_stop zenith_step azimuth_offset',
                                'LS1 rate_1 LS2 rate_2 number_of_datasets', ]

    # Specifications of the channel lines in the header
    licel_file_channel_format = 'active analog_photon laser_used number_of_datapoints 1 HV bin_width wavelength d1 d2 d3 d4 ADCbits number_of_shots discriminator ID'

    fix_zenith_angle = True
    skip_scan_overview_line = False  # Skip the 3d line containing azimuth_start, stop etc. Used to overcome a bug in some output files.

    def _read_rest_of_header(self, f):
        """ Read the third and fourth row of  of the header lines.

        The first two rows are read in the licel class.

        Parameters
        ----------
        f : file
           An open file-like object.

        Returns
        -------
        raw_info : dict
           A dictionary containing all parameters of the third and fourth line of the header.
        """
        raw_info = {}

        third_line = f.readline().decode()
        self.header_lines.append(third_line.strip())

        raw_info.update(self.match_lines(third_line, self.licel_file_header_format[2]))

        fourth_line = f.readline().decode()
        self.header_lines.append(fourth_line.strip())

        raw_info.update(self.match_lines(fourth_line, self.licel_file_header_format[3]))
        return raw_info

    def _assign_properties(self):
        """ Assign scanning-specific parameters found in the header as object properties."""
        super(ScanningFile, self)._assign_properties()
        self.azimuth_angle_raw = float(self.raw_info['azimuth_angle'])
        self.temperature = float(self.raw_info['temperature'])
        self.pressure = float(self.raw_info['pressure'])

        if not self.skip_scan_overview_line:
            self.azimuth_start_raw = float(self.raw_info['azimuth_start'])
            self.azimuth_stop_raw = float(self.raw_info['azimuth_stop'])
            self.azimuth_step = float(self.raw_info['azimuth_step'])
            self.zenith_start_raw = float(self.raw_info['zenith_start'])
            self.zenith_stop_raw = float(self.raw_info['zenith_stop'])
            self.zenith_step = float(self.raw_info['zenith_step'])
            self.azimuth_offset = float(self.raw_info['azimuth_offset'])
        else:
            self.azimuth_start_raw = np.nan
            self.azimuth_stop_raw = np.nan
            self.azimuth_step = np.nan
            self.zenith_start_raw = np.nan
            self.zenith_stop_raw = np.nan
            self.zenith_step = np.nan
            self.azimuth_offset = 0

        self.azimuth_angle = (self.azimuth_angle_raw + self.azimuth_offset) % 360
        self.azimuth_start = (self.azimuth_start_raw + self.azimuth_offset) % 360
        self.azimuth_stop = (self.azimuth_stop_raw + self.azimuth_offset) % 360

        if self.fix_zenith_angle:
            logger.debug('Fixing zenith start and zenith stop angles.')
            self.zenith_start = self._correct_zenith_angle(self.zenith_start_raw)
            self.zenith_stop = self._correct_zenith_angle(self.zenith_stop_raw)
        else:
            self.zenith_start = self.zenith_start_raw
            self.zenith_stop = self.zenith_stop_raw

        try:
            self.custom_info = self.raw_info['custom_info'].strip('"')
        except KeyError:
            self.custom_info = None

    def get_coordinates(self, channel_name):
        """
        Calculate the lat, lon, z coordinates for each measurement point.

        Parameters
        ----------
        channel_name : str
           The name of the channel. Only the channel object knows about the
           range resolution.

        Returns
        -------
        lat : array
           Latitude array
        lon : array
           Longitude array
        z : array
           Altitude array in meters
        """
        R_earth = 6378137  # Earth radius in meters

        # Shortcuts to make equations cleaner
        lat_center = self.latitude
        lon_center = self.longitude
        r = self.channels[channel_name].z
        azimuth = self.azimuth_angle
        zenith = self.zenith_angle

        # Convert all angles to radiants
        zenith_rad = np.deg2rad(zenith)[:, np.newaxis]
        azimuth_rad = np.deg2rad(azimuth)[:, np.newaxis]

        lat_center_rad = np.deg2rad(lat_center)
        lon_center_rad = np.deg2rad(lon_center)

        # Generate the mesh
        R, Zeniths = np.meshgrid(r, zenith_rad)
        R_ground = R * np.sin(Zeniths)
        Z = R * np.cos(Zeniths)

        # Equations from https://www.movable-type.co.uk/scripts/latlong.html
        delta = R_ground / R_earth
        lat_out_rad = np.arcsin(np.sin(lat_center_rad) * np.cos(delta)
                                + np.cos(lat_center_rad) * np.sin(delta) * np.cos(azimuth_rad))
        lon_out_rad = lon_center_rad + np.arctan2(np.sin(azimuth_rad) * np.sin(delta) * np.cos(lat_center_rad),
                                                  np.cos(delta) - np.sin(lat_center_rad) * np.sin(lat_out_rad))

        # Convert back to degrees
        lat_out = np.rad2deg(lat_out_rad)
        lon_out = np.rad2deg(lon_out_rad)

        return lat_out, lon_out, Z


class ScanningChannel(LicelChannel):
    """ A class representing measurements of a specific lidar channel, during a scanning measurement. """

    def __init__(self):
        super(ScanningChannel, self).__init__()

        self.azimuth_start = None
        self.azimuth_stop = None
        self.azimuth_step = None
        self.zenith_start = None
        self.zenith_stop = None
        self.zenith_step = None
        self.azimuth_offset = None
        self.zenith_angles = []
        self.azimuth_angles = []
        self.temperature = []
        self.pressure = []

    def append_file(self, current_file, file_channel):
        """ Keep track of scanning-specific variable properties of each file. """
        super(ScanningChannel, self).append_file(current_file, file_channel)
        self.zenith_angles.append(current_file.zenith_angle)
        self.azimuth_angles.append(current_file.azimuth_angle)
        self.temperature.append(current_file.temperature)
        self.pressure.append(current_file.pressure)

    def _assign_properties(self, current_file, file_channel):
        """ Assign scanning-specific properties as object properties. Check that these are unique,
        i.e. that all files belong to the same measurements set.

        Parameters
        ----------
        current_file : ScanningFile object
           A ScanningFile object being imported
        file_channel : LicelChannelData object
           A specific LicelChannelData object holding data found in the file.
        """
        super(ScanningChannel, self)._assign_properties(current_file, file_channel)
        self._assign_unique_property('azimuth_start', current_file.azimuth_start)
        self._assign_unique_property('azimuth_stop', current_file.azimuth_stop)
        self._assign_unique_property('azimuth_start_raw', current_file.azimuth_start_raw)
        self._assign_unique_property('azimuth_stop_raw', current_file.azimuth_stop_raw)
        self._assign_unique_property('azimuth_step', current_file.azimuth_step)
        self._assign_unique_property('zenith_start', current_file.zenith_start)
        self._assign_unique_property('zenith_stop', current_file.zenith_stop)
        self._assign_unique_property('zenith_start_raw', current_file.zenith_start_raw)
        self._assign_unique_property('zenith_stop_raw', current_file.zenith_stop_raw)
        self._assign_unique_property('zenith_step', current_file.zenith_step)

    def plot_ppi(self, figsize=(8, 4), signal_type='rc', z_min=0., z_max=12000., show_plot=True,
                 cmap=plt.cm.jet, title=None, vmin=0, vmax=1.3 * 10 ** 7, mask_noise=True, noise_threshold=1.):
        """
        Plot a vertical project of channel data.

        Parameters
        ----------
        figsize : tuple
           (width, height) of the output figure (inches)
        signal_type : str
           If 'rc', the range corrected signal is ploted. Else, the raw signals are used.
        z_min : float
           Minimum z range
        z_max : float
           Maximum z range
        show_plot : bool
           If True, the show_plot command is run.
        cmap : cmap
           An instance of a matplotlib colormap to use.
        z0 : float
           The ground-level altitude. If provided the plot shows altitude above sea level.
        title : str
           Optional title for the plot.
        vmin : float
           Minimum value for the color scale.
        vmax : float
           Maximum value for the color scale.
        mask_noise : bool
           If True, remove noisy bins.
        noise_threshold : int
           Threshold to use in the noise masking routine.
        """
        fig = plt.figure(figsize=figsize)
        ax1 = fig.add_subplot(111)

        self.draw_ppi(ax1, cmap=cmap, signal_type=signal_type, z_min=z_min, z_max=z_max, vmin=vmin, vmax=vmax,
                      mask_noise=mask_noise, noise_threshold=noise_threshold)

        if title:
            ax1.set_title(title)
        else:
            ax1.set_title("PPI scan")

        if show_plot:
            plt.show()

    def plot_rhi(self, figsize=(8, 4), signal_type='rc', z_min=0., z_max=12000., show_plot=True,
                 cmap=plt.cm.jet, title=None, vmin=0, vmax=1.3 * 10 ** 7, mask_noise=True, noise_threshold=1.):
        """
        Plot a vertical project of channel data.

        Parameters
        ----------
        figsize : tuple
           (width, height) of the output figure (inches)
        signal_type : str
           If 'rc', the range corrected signal is ploted. Else, the raw signals are used.
        z_min : float
           Minimum z range
        z_max : float
           Maximum z range
        show_plot : bool
           If True, the show_plot command is run.
        cmap : cmap
           An instance of a matplotlib colormap to use.
        z0 : float
           The ground-level altitude. If provided the plot shows altitude above sea level.
        title : str
           Optional title for the plot.
        vmin : float
           Minimum value for the color scale.
        vmax : float
           Maximum value for the color scale.
        mask_noise : bool
           If True, remove noisy bins.
        noise_threshold : int
           Threshold to use in the noise masking routine.
        """
        fig = plt.figure(figsize=figsize)
        ax1 = fig.add_subplot(111)

        projection_angle = self.draw_rhi(ax1, cmap=cmap, signal_type=signal_type, z_min=z_min, z_max=z_max, vmin=vmin,
                                         vmax=vmax,
                                         mask_noise=mask_noise, noise_threshold=noise_threshold)

        if title:
            ax1.set_title(title)
        else:
            ax1.set_title("RHI scan ({0}$^\circ$)".format(projection_angle))

        if show_plot:
            plt.show()

    def plot_scan(self, figsize=(8, 4), signal_type='rc', z_min=0., z_max=12000., show_plot=True,
                  cmap=plt.cm.jet, vmin=0, vmax=1.3 * 10 ** 7, mask_noise=True, noise_threshold=1., cb_format='%.0e',
                  box=False, grid=(1, 4), ax1_position=(0, 0), ax1_span=2, ax2_position=(0, 2), ax2_span=2):
        """
        Plot data as RHI and PPI scans.

        Parameters
        ----------
        figsize : tuple
           (width, height) of the output figure (inches)
        signal_type : str
           If 'rc', the range corrected signal is ploted. Else, the raw signals are used.
        z_min : float
           Minimum z range
        z_max : float
           Maximum z range
        show_plot : bool
           If True, the show_plot command is run.
        cmap : cmap
           An instance of a matplotlib colormap to use.
        z0 : float
           The ground-level altitude. If provided the plot shows altitude above sea level.
        title : str
           Optional title for the plot.
        vmin : float
           Minimum value for the color scale.
        vmax : float
           Maximum value for the color scale.
        mask_noise : bool
           If True, remove noisy bins.
        noise_threshold : int
           Threshold to use in the noise masking routine.
        """
        fig = plt.figure(figsize=figsize)

        ax1 = plt.subplot2grid(grid, ax1_position, colspan=ax1_span)
        ax2 = plt.subplot2grid(grid, ax2_position, colspan=ax2_span)

        self.draw_ppi(ax1, cmap=cmap, signal_type=signal_type, z_min=z_min, z_max=z_max, vmin=vmin, vmax=vmax,
                      mask_noise=mask_noise, noise_threshold=noise_threshold, add_colorbar=False, cb_format=cb_format,
                      box=box)

        projection_angle = self.draw_rhi(ax2, cmap=cmap, signal_type=signal_type, z_min=z_min, z_max=z_max, vmin=vmin,
                                         vmax=vmax,
                                         mask_noise=mask_noise, noise_threshold=noise_threshold, cb_format=cb_format,
                                         box=box)

        fig.suptitle("Channel {0}: {1} - {2}".format(self.name,
                                                     self.start_time.strftime('%Y%m%dT%H%M'),
                                                     self.stop_time.strftime('%Y%m%dT%H%M')))

        ax1.set_title('PPI')
        ax2.set_title("RHI ({0:.1f}$^\circ$)".format(projection_angle))

        plt.tight_layout()
        plt.subplots_adjust(top=0.85)

        if show_plot:
            plt.show()

    def draw_ppi(self, ax1, cmap=plt.cm.jet, signal_type='rc',
                 z_min=0, z_max=12000., add_colorbar=True, cmap_label='a.u.', cb_format=None,
                 vmin=0, vmax=1.3 * 10 ** 7, mask_noise=True, noise_threshold=1., first_signal_bin=0, box=False):
        """
        Draw channel data as a PPI plot.

        Parameters
        ----------
        ax1 : axis object
           The axis object to draw.
        x : array
           X axis coordinates
        y : array
           Y axis coordinates
        cmap : cmap
           An instance of a matplotlib colormap to use.
        signal_type : str
           If 'rc', the range corrected signal is ploted. Else, the raw signals are used.
        z_min : float
           Minimum z range
        z_max : float
           Maximum z range
        add_colorbar : bool
           If True, a colorbar will be added to the plot.
        cmap_label : str
           Label for the colorbar. Ignored if add_colorbar is False.
        cb_format : str
           Colorbar tick format string.
        vmin : float
           Minimum value for the color scale.
        vmax : float
           Maximum value for the color scale.
        mask_noise : bool
           If True, remove noisy bins.
        noise_threshold : int
           Threshold to use in the noise masking routine.
        first_signal_bin : int
           First signal bin. Can be used to fix analog bin shift of Licel channels.
        """
        x, y = self._polar_to_ground(self.z / 1000., self.azimuth_angles, self.zenith_angles)

        self.draw_projection(ax1, x, y, cmap=cmap, signal_type=signal_type,
                             z_min=z_min, z_max=z_max, add_colorbar=add_colorbar, cmap_label=cmap_label,
                             cb_format=cb_format, vmin=vmin, vmax=vmax, mask_noise=mask_noise,
                             noise_threshold=noise_threshold, first_signal_bin=first_signal_bin)

        if box:
            ax1.set_xlim(-z_max / 1000., z_max / 1000.)
            ax1.set_ylim(-z_max / 1000., z_max / 1000.)

        ax1.set_ylabel('South-North (km)')
        ax1.set_xlabel('West-East (km)')

    def draw_rhi(self, ax1, cmap=plt.cm.jet, signal_type='rc',
                 z_min=0, z_max=12000., add_colorbar=True, cmap_label='a.u.', cb_format=None,
                 vmin=0, vmax=1.3 * 10 ** 7, mask_noise=True, noise_threshold=1., first_signal_bin=0, box=False):
        """
        Draw channel data as a PPI plot.

        Parameters
        ----------
        ax1 : axis object
           The axis object to draw.
        x : array
           X axis coordinates
        y : array
           Y axis coordinates
        cmap : cmap
           An instance of a matplotlib colormap to use.
        signal_type : str
           If 'rc', the range corrected signal is plotted. Else, the raw signals are used.
        z_min : float
           Minimum z range
        z_max : float
           Maximum z range
        add_colorbar : bool
           If True, a colorbar will be added to the plot.
        cmap_label : str
           Label for the colorbar. Ignored if add_colorbar is False.
        cb_format : str
           Colorbar tick format string.
        vmin : float
           Minimum value for the color scale.
        vmax : float
           Maximum value for the color scale.
        mask_noise : bool
           If True, remove noisy bins.
        noise_threshold : int
           Threshold to use in the noise masking routine.
        first_signal_bin : int
           First signal bin. Can be used to fix analog bin shift of Licel channels.
        """
        projection_angle = np.mean(self.azimuth_angles)
        x, y = self._polar_to_cross_section(self.z / 1000., self.azimuth_angles, self.zenith_angles, projection_angle)

        self.draw_projection(ax1, x, y, cmap=cmap, signal_type=signal_type,
                             z_min=z_min, z_max=z_max, add_colorbar=add_colorbar, cmap_label=cmap_label,
                             cb_format=cb_format, vmin=vmin, vmax=vmax, mask_noise=mask_noise,
                             noise_threshold=noise_threshold, first_signal_bin=first_signal_bin)

        padding = 0.5  # km

        if box:
            ax1.set_xlim(-z_max / 1000. - padding, z_max / 1000. + padding)
            ax1.set_ylim(-padding, z_max / 1000. + padding)

        ax1.set_xlabel('Distance (km)')
        ax1.set_ylabel('Height a.l. (km)')

        return projection_angle

    def draw_projection(self, ax1, x, y, cmap=plt.cm.jet, signal_type='rc',
                        z_min=0, z_max=12000., add_colorbar=True, cmap_label='a.u.', cb_format=None,
                        vmin=0, vmax=1.3 * 10 ** 7, mask_noise=True, noise_threshold=1.,
                        first_signal_bin=0):
        """
        Draw channel data as a PPI plot.

        Parameters
        ----------
        ax1 : axis object
           The axis object to draw.
        x : array
           X axis coordinates
        y : array
           Y axis coordiantes
        cmap : cmap
           An instance of a matplotlib colormap to use.
        signal_type : str
           If 'rc', the range corrected signal is ploted. Else, the raw signals are used.
        z_min : float
           Minimum z range
        z_max : float
           Maximum z range
        add_colorbar : bool
           If True, a colorbar will be added to the plot.
        cmap_label : str
           Label for the colorbar. Ignored if add_colorbar is False.
        cb_format : str
           Colorbar tick format string.
        vmin : float
           Minimum value for the color scale.
        vmax : float
           Maximum value for the color scale.
        mask_noise : bool
           If True, remove noisy bins.
        noise_threshold : int
           Threshold to use in the noise masking routine.
        first_signal_bin : int
           First signal bin. Can be used to fix analog bin shift of Licel channels.
        """
        if signal_type == 'rc':
            if len(self.rc) == 0:
                self.calculate_rc()
            data = self.rc
        else:
            data = self.matrix

        if mask_noise:
            mask = self.noise_mask(threshold=noise_threshold)
            data = np.ma.masked_where(mask, data)

        z_min_idx = self._index_at_height(z_min)
        z_max_idx = self._index_at_height(z_max)

        data_min_idx = z_min_idx + first_signal_bin
        data_max_idx = z_max_idx + first_signal_bin

        im1 = ax1.pcolormesh(x[:, z_min_idx:z_max_idx], y[:, z_min_idx:z_max_idx], data[:, data_min_idx:data_max_idx],
                             cmap=cmap, vmin=vmin, vmax=vmax)

        ax1.set(adjustable='box', aspect='equal')

        if add_colorbar:
            if cb_format:
                cb1 = plt.colorbar(im1, format=cb_format)
            else:
                cb1 = plt.colorbar(im1)
            cb1.ax.set_ylabel(cmap_label)

            # Make the ticks of the colorbar smaller, two points smaller than the default font size
            cb_font_size = mpl.rcParams['font.size'] - 2
            for ticklabels in cb1.ax.get_yticklabels():
                ticklabels.set_fontsize(cb_font_size)
            cb1.ax.yaxis.get_offset_text().set_fontsize(cb_font_size)

        # Centered axis in center: https://stackoverflow.com/a/31558968
        # Move left y-axis and bottim x-axis to centre, passing through (0,0)
        # ax1.spines['left'].set_position('center')
        # ax1.spines['bottom'].set_position('center')
        #
        # # Eliminate upper and right axes
        # ax1.spines['right'].set_color('none')
        # ax1.spines['top'].set_color('none')
        #
        # # Show ticks in the left and lower axes only
        # ax1.xaxis.set_ticks_position('bottom')
        # ax1.yaxis.set_ticks_position('left')

    @staticmethod
    def _polar_to_ground(z, azimuth, zenith):
        """
        Convert polar coordinates to cartesian project for a PPI scan

        Parameters
        ----------
        z : array
           Distance array in meters
        azimuth : list
           List of profile azimuth angles in degrees
        zenith : list
           List of profile zenith angles in degrees

        Returns
        -------
        x : array
           X axis in meters
        y : array
           Y axis in meters
        """
        # Generate the mesh
        zenith_rad = np.deg2rad(zenith)
        azimuth_rad = np.deg2rad(azimuth)

        Z, Zeniths = np.meshgrid(z, zenith_rad)
        Z_ground = Z * np.sin(Zeniths)

        x = Z_ground * np.sin(azimuth_rad)[:, np.newaxis]
        y = Z_ground * np.cos(azimuth_rad)[:, np.newaxis]

        return x, y

    @staticmethod
    def _polar_to_cross_section(z, azimuth, zenith, cross_section_azimuth):
        """
        Convert polar coordinates to cartesian project for a PPI scan

        Parameters
        ----------
        z : array
           Distance array in meters
        azimuth : list
           List of profile azimuth angles in degrees
        zenith : list
           List of profile zenith angles in degrees
        cross_section_azimuth : float
           Azimuth angle of plane in degrees

        Returns
        -------
        x : array
           X axis in meters
        y : array
           Y axis in meters
        """

        zenith_rad = np.deg2rad(zenith)

        # The angle between measurements and the cross section plance
        azimuth_difference_rad = np.deg2rad(azimuth) - np.deg2rad(cross_section_azimuth)

        # Generate the mesh
        Z, Azimuth_differences = np.meshgrid(z, azimuth_difference_rad)

        x = Z * np.sin(zenith_rad)[:, np.newaxis] * np.cos(Azimuth_differences)
        y = Z * np.cos(zenith_rad)[:, np.newaxis]

        return x, y

    def get_coordinates(self,):
        """
        Calculate the lat, lon, z coordinates for each measurement point.

        Returns
        -------
        lat : array
           Latitude array
        lon : array
           Longitude array
        z : array
           Altitude array in meters
        """
        R_earth = 6378137  # Earth radius in meters

        # Shortcuts to make equations cleaner
        lat_center = self.latitude
        lon_center = self.longitude
        r = self.z
        azimuth = self.azimuth_angles
        zenith = self.zenith_angles

        # Convert all angles to radiants
        zenith_rad = np.deg2rad(zenith)[:, np.newaxis]
        azimuth_rad = np.deg2rad(azimuth)[:, np.newaxis]

        lat_center_rad = np.deg2rad(lat_center)
        lon_center_rad = np.deg2rad(lon_center)

        # Generate the mesh
        R, Zeniths = np.meshgrid(r, zenith_rad)
        R_ground = R * np.sin(Zeniths)
        Z = R * np.cos(Zeniths)

        # Equations from https://www.movable-type.co.uk/scripts/latlong.html
        delta = R_ground / R_earth
        lat_out_rad = np.arcsin(np.sin(lat_center_rad) * np.cos(delta)
                                + np.cos(lat_center_rad) * np.sin(delta) * np.cos(azimuth_rad))
        lon_out_rad = lon_center_rad + np.arctan2(np.sin(azimuth_rad) * np.sin(delta) * np.cos(lat_center_rad),
                                                  np.cos(delta) - np.sin(lat_center_rad) * np.sin(lat_out_rad))

        # Convert back to degrees
        lat_out = np.rad2deg(lat_out_rad)
        lon_out = np.rad2deg(lon_out_rad)

        return lat_out, lon_out, Z


class ScanningFileMissingLine(ScanningFile):
    skip_scan_overview_line = True


class ScanningLidarMeasurement(LicelLidarMeasurement):
    """ A class representing a scanning measurement set.

    It useses `ScanningFile` and `ScanningChannel` classes for handling the data.
    """
    file_class = ScanningFile
    channel_class = ScanningChannel
    photodiode_class = PhotodiodeChannel


class FixedPointingFile(LicelFile):
    """ Raymetrics is using a custom version of licel file format to store
    vertical lidar measurements.

    `azimuth_angle`
        The angle from North

    `temperature`
        Ambient temperature (degrees C)

    `pressure`
        Ambient pressure (hPa)
    """
    # Specifications of the header lines.
    licel_file_header_format = ['filename',
                                'start_date start_time end_date end_time altitude longitude latitude zenith_angle azimuth_angle temperature pressure custom_info',
                                # Appart from Site that is read manually
                                'LS1 rate_1 LS2 rate_2 number_of_datasets', ]

    fix_zenith_angle = True

    def __init__(self, file_path, use_id_as_name=False, get_name_by_order=False, licel_timezone="UTC", import_now=True, azimuth_offset=0):
        self.azimuth_offset = azimuth_offset

        super().__init__(file_path, use_id_as_name=use_id_as_name, get_name_by_order=get_name_by_order,
                         licel_timezone=licel_timezone, import_now=import_now)

    def _assign_properties(self):
        """ Assign scanning-specific parameters found in the header as object properties."""
        super(FixedPointingFile, self)._assign_properties()

        self.temperature = float(self.raw_info['temperature'])
        self.pressure = float(self.raw_info['pressure'])
        self.azimuth_angle_raw = float(self.raw_info['azimuth_angle'])

        azimuth_offset = getattr(self, 'azimuth_offset', None)

        # Add a option to manually define an azimuth offset, but still work if you don't
        if azimuth_offset is not None:
            self.azimuth_angle = (self.azimuth_angle_raw + azimuth_offset) % 360
        else:
            self.azimuth_angle = self.azimuth_angle_raw

        try:
            self.custom_info = self.raw_info['custom_info'].strip('"')
        except KeyError:
            self.custom_info = None


class FixedPointingChannel(LicelChannel):
    """ A class representing measurements of a specific lidar channel, during a fixed pointing measurement. """

    def __init__(self):
        super(FixedPointingChannel, self).__init__()
        self.zenith_angles = []
        self.azimuth_angles = []
        self.temperature = []
        self.pressure = []

    def append_file(self, current_file, file_channel):
        """ Keep track of scanning-specific variable properties of each file. """
        super(FixedPointingChannel, self).append_file(current_file, file_channel)
        self.zenith_angles.append(current_file.zenith_angle)
        self.azimuth_angles.append(current_file.azimuth_angle)
        self.temperature.append(current_file.temperature)
        self.pressure.append(current_file.pressure)


class FixedPointingLidarMeasurement(LicelLidarMeasurement):
    file_class = FixedPointingFile
    channel_class = FixedPointingChannel



from datetime import datetime, date
import sys
import logging
from pathlib import Path
import xarray as xr

import matplotlib.pyplot as plt
from gfatpy.lidar.scc.plot.utils import plot_angstrom
from gfatpy.lidar.scc.plot.retrieval import angstrom_exponent

from gfatpy.utils.io import unzip_file
from gfatpy.lidar.scc.plot.product import SCC_elpp, SCC_elda, SCC_raw

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel(logging.INFO)


class SCC_zipfile:
    """A simple class to manage SCC products."""

    def __init__(self, path: Path):
        """Initialize SCC_zipfile class.

        Args:
            path (Path): Path of SCC zip file.
        """
        self.path = self.check_path(path)
        self.type, self.measurementID = self._get_info(self.path)
        self.unzipped_path = self.extract_zip()
        self.products = self.get_products()
        self.hoi_configuration_ID = self.get_hoi_configuration_ID()
        self.date = self.get_date_from_measurementID(measurementID=self.measurementID)

    def check_path(self, path: Path) -> Path:
        """Check if the path exists and is a zip file.

        Args:
            path (Path): Path of SCC zip file.

        Raises:
            ValueError: Filepath does not exist.
            ValueError: Filepath is not a zip file.

        Returns:
            Path: Path of SCC zip file.
        """
        if not path.exists():
            raise ValueError("Filepath does not exist.")
        # check zip
        if not path.suffix.endswith("zip"):
            raise ValueError(f"Filepath {path} is not a zip file.")
        return path

    def _get_info(self, path: Path) -> tuple[str, str]:
        """Get info from SCC zip file.

        Args:
            path (Path): Path of SCC zip file.

        Returns:
            tuple[str: str]: SCC product type (scc type as elpp, elda, ...) and measurementID.
        """
        # split pathfilename elpp_20230706gra0200.zip
        pathfilename = path.name
        type, filename = pathfilename.split("_")
        if type == "optical":
            type = "elda"
        if type == "preprocessed":
            type = "elpp"
        return type, filename.split(".")[0]

    def extract_zip(
        self,
        pattern_or_list: str | list[str] = rf"\.nc$",
        destination: Path | None = None,
    ) -> Path | None:
        """Extract zip file.

        Args:
            pattern (str, optional): Pattern to search in zip file. Defaults to f"*.nc".
            destination (Path | None, optional): Destination path to extract zip file. Defaults to None means current working directory.

        Returns:
            Path | None: Path of unzipped folder or None if nothing is extracted.
        """
        return unzip_file(
            self.path, pattern_or_list=pattern_or_list, destination=destination
        )

    def get_date_from_measurementID(self, measurementID: str) -> date:
        """Get date from measurementID.

        Args:
            measurementID (str): measurementID as 20230706gra0200

        Returns:
            datetime: datetime object.
        """
        return datetime.strptime(measurementID[0:8], "%Y%m%d").date()

    def get_products(self) -> list[SCC_raw | SCC_elpp | SCC_elda]:
        """Get products from unzipped folder.

        Returns:
            list[SCC_elpp | SCC_elda]: List of SCC_elpp or SCC_elda objects.
        """

        if self.type == "raw":
            return [SCC_raw(file_) for file_ in self.unzipped_path.rglob(f"*.nc")]

        if self.type == "elpp":
            return [
                SCC_elpp(file_)
                for file_ in self.unzipped_path.rglob(f"*{self.type}*.nc")
            ]

        if self.type == "elda":
            return [
                SCC_elda(file_)
                for file_ in self.unzipped_path.rglob(f"*{self.type}*.nc")
            ]

    def angstrom_exponent(self) -> dict[str : xr.DataArray]:
        """Retrieve angstrom exponent from elda products.

        Returns:
            dict[str: xr.DataArray]: Dictionary with angstrom exponents.
        """
        dict_angstrom_exponent = {}
        for product1 in self.products:
            if product1.earlinet_product_type[0] != "b":
                continue
            for product2 in self.products:
                if product2.earlinet_product_type[0] != "b":
                    continue
                if product1.wavelength >= product2.wavelength:
                    continue
                name_ = f"{product1.wavelength}-{product2.wavelength}{product1.string_ID[-1]}"
                dict_angstrom_exponent[name_] = angstrom_exponent(product1, product2)
        return dict_angstrom_exponent

    def get_hoi_configuration_ID(self) -> int | None:
        """Get hoi_configuration_ID from global attributes.

        Returns:
            int: hoi_configuration_ID.
        """
        # get hoi_configuration_ID from global attributes

        if self.type == "raw":
            return None

        with xr.open_dataset(self.products[0].path) as ds_:
            return ds_.attrs["hoi_configuration_ID"]

    def plot_raw(
        self,
        output_dir: Path | None = None,
        dpi: int = 300,
        raw_limits: tuple[float, float] | None = None,
        range_limits: tuple[float, float] | None = None,
        combine_channels: bool = False,
        savefig: bool = True,
    ) -> list[Path | None]:
        """Plot raw products.

        Args:
            output_dir (Path | None, optional): Path to save plots. Defaults to None means current working directory.
            dpi (int, optional): Dots per inch. Defaults to 300.

        Returns:
            list[Path]: Paths of the plots.
        """
        return [
            product.plot(
                output_dir=output_dir,
                dpi=dpi,
                raw_limits=raw_limits,
                range_limits=range_limits,
                savefig=savefig,
                combine_channels=combine_channels,
            )[-1]
            for product in self.products
        ]

    def plot_elpp(
        self,
        output_dir: Path | None = None,
        dpi: int = 300,
        attenuated_backscatter_limits: tuple[float, float] | None = None,
        normalization_range: tuple[float, float] | None = None,
        range_limits: tuple[float, float] | None = None,
        combine_channels: bool = False,
        savefig: bool = True,
    ) -> list[Path | None]:
        """Plot elpp products.

        Args:
            output_dir (Path | None, optional): Path to save plots. Defaults to None means current working directory.
            dpi (int, optional): Dots per inch. Defaults to 300.

        Returns:
            list[Path]: Paths of the plots.
        """
        return [
            product.rayleight_fit(
                output_dir=output_dir,
                dpi=dpi,
                attenuated_backscatter_limits=attenuated_backscatter_limits,
                normalization_range=normalization_range,
                range_limits=range_limits,
                savefig=savefig,
                combine_channels=combine_channels,
            )[-1]
            for product in self.products
        ]

    def plot_elda(
        self,
        output_dir: Path | None = None,
        dpi: int = 300,
        range_limits: tuple[float, float] | None = None,
        backscatter_limits: tuple[float, float] | None = None,
        extinction_limits: tuple[float, float] | None = None,
        lidar_ratio_limits: tuple[float, float] | None = None,
        ae_limits: tuple[float, float] | None = None,
        depo_limits: tuple[float, float] | None = None,
    ) -> Path:

        """Plot elda products

        Args:
            output_dir (Path | None, optional): Path to save plots. Defaults to None means current working directory.
            dpi (int, optional): Dots per inch. Defaults to 300.

        Returns:
            Path: Path of the png image.
        """
        fig, (
            ax_backscatter,
            ax_angstrom,
            ax_depo,
            ax_lidar_ratio,
            ax_extinction,
        ) = plt.subplots(nrows=1, ncols=5, figsize=[15, 7.5], sharey=True)
        for product in self.products:
            optical_product = product.earlinet_product_type
            if optical_product[0] == "b":
                ax_backscatter = product.plot_backscatter(
                    ax=ax_backscatter,
                    backscatter_limits=backscatter_limits,
                    range_limits=range_limits,
                )
                if "dr" in product.string_ID:
                    ax_depo = product.plot_depo(
                        ax=ax_depo, depo_limits=depo_limits, range_limits=range_limits
                    )
            elif optical_product[0] == "e":
                ax_extinction = product.plot_extinction(
                    ax=ax_extinction,
                    extinction_limits=extinction_limits,
                    range_limits=range_limits,
                )
                if "lr" in product.string_ID:
                    ax_lidar_ratio = product.plot_lidar_ratio(
                        ax=ax_lidar_ratio,
                        lidar_ratio_limits=lidar_ratio_limits,
                        range_limits=range_limits,
                    )

        ax_angstrom = plot_angstrom(
            self.angstrom_exponent(),
            ax=ax_angstrom,
            ae_limits=ae_limits,
            range_limits=range_limits,
        )

        # Fix y label. All empty but the first with "Range, [km]"
        for ax in [ax_backscatter, ax_angstrom, ax_depo, ax_lidar_ratio, ax_extinction]:
            ax.set_ylabel("")
            ax.set_title("")
            ax.legend(fontsize=10, loc="upper right")
            ax.grid()
            ax.minorticks_on()
        ax_backscatter.set_ylabel("Range, [km]")
        fig.suptitle(
            f"SCC config: {self.products[0].hoi_configuration_ID} | Measurement ID: {self.measurementID} | Averaged Time: {self.products[0].datetime_ini.strftime('%H:%M')} - {self.products[0].datetime_end.strftime('%H:%M')}"
        )
        fig.tight_layout()
        # save
        if isinstance(output_dir, Path):
            output_dir.mkdir(parents=True, exist_ok=True)
        elif output_dir is None:
            output_dir = Path.cwd()
        outputpath = output_dir.joinpath(self.path.name.replace(".zip", ".png"))
        fig.savefig(outputpath, dpi=dpi)
        plt.close(fig)
        return outputpath

    def plot(
        self,
        output_dir: Path | None = None,
        dpi: int = 300,
        range_limits: tuple[float, float] = (0, 25),
        raw_limits: tuple[float, float] | None = None,
        attenuated_backscatter_limits: tuple[float, float] | None = None,
        backscatter_limits: tuple[float, float] | None = None,
        extinction_limits: tuple[float, float] | None = None,
        lidar_ratio_limits: tuple[float, float] | None = None,
        ae_limits: tuple[float, float] | None = None,
        depo_limits: tuple[float, float] | None = None,
        elpp_combine_channels: bool = False,
        elpp_normalization_range: tuple[float, float] = (6, 7),
        savefig: bool = True,
    ) -> Path | list[Path]:
        """Plot content of SCC zip file.

        Args:
            output_dir (Path | None, optional): Directory to save plots. Defaults to None means current working directory.
            dpi (int, optional): Dots per inch. Defaults to 300.
            range_limits (tuple, optional): Set range limit in km. Defaults to None means (0,14).
            raw_limits (tuple, optional): Set raw limit in arbitrary units. Defaults to None means autofix.
            backscatter_limits (tuple, optional): Set backscatter limit in Mm^-1 sr^-1. Defaults to None means autofix.
            extinction_limits (tuple, optional): Set extinction limit in Mm^-1. Defaults to None means autofix.
            lidar_ratio_limits (tuple, optional): Set lidar ratio limit in sr. Defaults to None means values from gfatpy.lidar.plot.info.yml
            ae_limits (tuple, optional): Set angstrom exponent limit. Defaults to None means values from gfatpy.lidar.plot.info.yml
            depo_limits (tuple, optional): Set depolarization limit. Defaults to None means values from gfatpy.lidar.plot.info.yml

        Returns:
            Path | list[Path]: List of plots or single plot.
        """

        if self.type == "raw":
            return self.plot_raw(
                output_dir=output_dir,
                dpi=dpi,
                raw_limits=raw_limits,
                range_limits=range_limits,
                combine_channels=True,
                savefig=savefig,
            )

        if self.type == "elpp":
            return self.plot_elpp(
                output_dir=output_dir,
                dpi=dpi,
                attenuated_backscatter_limits=attenuated_backscatter_limits,
                normalization_range=elpp_normalization_range,
                range_limits=range_limits,
                combine_channels=elpp_combine_channels,
                savefig=savefig,
            )

        elif self.type == "elda":
            return self.plot_elda(
                output_dir=output_dir,
                dpi=dpi,
                range_limits=range_limits,
                backscatter_limits=backscatter_limits,
                extinction_limits=extinction_limits,
                lidar_ratio_limits=lidar_ratio_limits,
                ae_limits=ae_limits,
                depo_limits=depo_limits,
            )



from loguru import logger
import numpy as np
from datetime import datetime
from pathlib import Path
import pandas as pd
import xarray as xr

from .gamma_distribution import BinnedPSD

def getEdges():

    pars_class = np.zeros(shape=(32, 2))
    bin_edges = np.zeros(shape=(33, 1))

    # pars_class[:,0] : Center of Class [mm]
    # pars_class[:,1] : Width of Class [mm]
    pars_class[0:10, 1] = 0.125
    pars_class[10:15, 1] = 0.250
    pars_class[15:20, 1] = 0.500
    pars_class[20:25, 1] = 1.0
    pars_class[25:30, 1] = 2.0
    pars_class[30:32, 1] = 3.0

    j = 0
    pars_class[0, 0] = 0.062
    for i in range(1, 32):
        if (
            i < 10
            or (i > 10 and i < 15)
            or (i > 15 and i < 20)
            or (i > 20 and i < 25)
            or (i > 25 and i < 30)
            or (i > 30)
        ):
            pars_class[i, 0] = pars_class[i - 1, 0] + pars_class[i, 1]

        const = [0.188, 0.375, 0.75, 1.5, 2.5]
        if i == 10 or i == 15 or i == 20 or i == 25 or i == 30:
            pars_class[i, 0] = pars_class[i - 1, 0] + const[j]
            j = j + 1

        # print pars_class[i,0]
        bin_edges[i + 1, 0] = pars_class[i, 0] + pars_class[i, 1] / 2

    bin_edges[0, 0] = 0.0
    bin_edges[1, 0] = pars_class[0, 0] + pars_class[0, 1] / 2

    return bin_edges


def retrieve_dBZe_from_parsivel(    
    parsFilePath: Path | str,
    freq: float,
    surf_temp: float,
    startDate: datetime,
    stopDate: datetime,
    diameter_resolution: float = 0.01,
) -> xr.Dataset:
    """Converts Parsivel data to reflectivity.

    Args:
        parsFilePath (Path | str): disdrometer file with extension *.nc
        freq (float): Frequency [GHz] of the reflectivity
        surf_temp (float): Surface temperature [C]
        startDate (datetime.date): Start date of the period to be analyzed
        stopDate (datetime.date): Stop date of the period to be analyzed

    Raises:
        FileNotFoundError: If the file does not exist. 

    Returns:
        xr.Dataset: Reflectivity dataset
    """    

    freqstr = "%3.1f" % freq

    # reding meang volume diameter from parsivel
    if isinstance(parsFilePath, str):
        parsFilePath = Path(parsFilePath)
    
    if not parsFilePath.exists():
        raise FileNotFoundError(f"File {parsFilePath} does not exist")

    pasrDS = xr.open_dataset(parsFilePath)

    # Load scattering database
    if surf_temp < 5:
        scattabstr = Path(__file__).parent / "scattering_databases" / f"0.C_{freqstr}GHz.csv"
    elif 5 <= surf_temp < 15:        
        scattabstr = Path(__file__).parent / "scattering_databases" / f"10C_{freqstr}GHz.csv"
    else:
        scattabstr = Path(__file__).parent / "scattering_databases" / f"20C_{freqstr}GHz.csv"
    df = pd.read_csv(scattabstr)

    diameter = "diameter[mm]"
    radarxs = "radarsx[mm2]"
    wavelength = "wavelength[mm]"
    temp = "T[k]"
    extxs = "extxs[mm2]"

    upscale_end = (len(df) + 1.0) / 100.0
    diameter_ups = np.arange(diameter_resolution, upscale_end, diameter_resolution)

    # constants
    T = df.loc[1, temp]
    wavelen = df.loc[1, wavelength]
    K2 = 0.93

    # integration constant
    int_const = wavelen**4 / ((np.pi) ** 5 * K2)

    #select time range    
    before=( pd.to_datetime(startDate,format='%Y-%m-%d %H:%M') < pd.to_datetime(pasrDS["time"].data))
    after = ( pd.to_datetime(stopDate,format='%Y-%m-%d %H:%M') > pd.to_datetime(pasrDS["time"].data))
    mask_time = before & after
                          
    if not any(mask_time):
        print('No rain in this user-provided period.')
        return
    else:
        parstime = pasrDS["time"].sel(time=slice(startDate, stopDate))
        N = pasrDS['droplet_number_concentration'].sel(time=slice(startDate, stopDate)).values
        breakpoint()
        N = 10**N
        #print(N)
        
    #calculating Ze using parsivel SD    
    Zpars = np.zeros(N.shape[0])    
    if freq > 20:        
        #Ze parsivel using T-matrix at "freq" GHz
        for i in range(N.shape[0]):
            PSD = BinnedPSD(getEdges(),N[i])                                        
            y = PSD(diameter_ups)*np.tile(df.loc[:,radarxs], reps=(PSD(diameter_ups).shape[1],1)).T
            Zpars[i] = int_const * y.sum()*diameter_resolution
            logger.info(f"Ze parsivel in Mie regime. Using T-matrix at {freqstr} GHz. Retrieval NOT VALIDATED.")
    else:
        #Ze parsivel in Rayleigh regime
        for i in range(N.shape[0]):
            PSD = BinnedPSD(getEdges(),N[i])
            d6 = PSD(diameter_ups)*df.loc[:,diameter]**6
            Zpars[i] = d6.sum()*diameter_resolution 
      
    parsZe = 10 * np.log10(Zpars)
    pars = xr.Dataset({'parsZe': (['time'],  parsZe)}, coords={'time': parstime})
    return pars


""" This is a class for experimenting with the new DIVA / EARLINET NetCDF file format.

In the long run, this should be places as a method in BaseLidarMeasurement class. For now it is kept
separately not to interfere with normal development.
"""
import netCDF4 as netcdf
import yaml
import datetime
import os
import numpy as np
import logging

import pytz

logger = logging.getLogger(__name__)


class DivaConverterMixin:

    def save_as_diva_netcdf(self, output_path, parameter_file):
        """ Save the current data in the 'draft' DIVA format. """

        with open(parameter_file, 'r') as f:
            parameters = yaml.load(f)

        global_parameters = parameters['global_parameters']  # Shortcut
        global_variables = parameters['global_variables']  # Shortcut
        channels = parameters['channels']

        iso_date = datetime.datetime.utcnow().strftime('%Y-%d-%mT%H:%M:%SZ')
        python_file_name = os.path.basename(__file__)

        with netcdf.Dataset(output_path, 'w', format="NETCDF4") as f:

            # Global attributes
            f.title = global_parameters['title']
            f.source = global_parameters['source']
            f.institution = global_parameters['institution']
            f.references = global_parameters['references']
            f.location = global_parameters['location']
            f.data_version = global_parameters['data_version']
            f.PI = global_parameters['PI_name']
            f.PI_email = global_parameters['PI_email']
            f.conversion_date = iso_date
            f.comment = global_parameters['comment']
            f.Conventions = global_parameters['Conventions']
            f.history = global_parameters['history'].format(date=iso_date, file=python_file_name)
            f.featureType = "timeSeriesProfile"

            # Top level dimensions
            f.createDimension('name_strlen', size=40)
            f.createDimension('nv', size=2)

            # Top level variables
            latitude = f.createVariable('latitude', datatype='f4')
            latitude.standard_name = 'latitude'
            latitude.long_name = 'system latitude'
            latitude.units = 'degrees_north'

            longitude = f.createVariable('longitude', datatype='f4')
            longitude.standard_name = 'longitude'
            longitude.long_name = 'system longitude'
            longitude.units = 'degrees_east'

            lidar_zenith_angle = f.createVariable('lidar_zenith_angle', datatype='f4')
            lidar_zenith_angle.standard_name = 'sensor_zenith_angle'
            lidar_zenith_angle.long_name = 'zenith angle of emitted laser'
            lidar_zenith_angle.units = 'degree'

            lidar_azimuth = f.createVariable('lidar_azimuth_angle', datatype='f4')
            lidar_azimuth.standard_name = 'sensor_azimuth_angle'
            lidar_azimuth.long_name = 'azimuth angle of emitted laser'
            lidar_azimuth.units = 'degree'
            lidar_azimuth.comment = 'Based on North. Optional'

            altitude = f.createVariable('altitude', datatype='f4')
            altitude.standard_name = 'altitude'
            altitude.long_name = 'system altitude'
            altitude.units = 'm'

            # Assign top-level variables
            latitude[:] = global_variables['latitude']
            longitude[:] = global_variables['longitude']
            lidar_zenith_angle[:] = global_variables['laser_pointing_angle']
            altitude[:] = global_variables['system_altitude']

            # Optional ancillary group
            ancillary = f.createGroup('ancillary')
            ancillary.featureType = "timeSeries"

            ancillary.createDimension('time', size=None)

            time = ancillary.createVariable('time', datatype='f8', dimensions=('time',))
            time.long_name = 'time'
            time.units = 'seconds since 1970-01-01 00:00'
            time.standard_name = 'time'

            temperature = ancillary.createVariable('air_temperature', datatype='f8', dimensions=('time',))
            temperature.long_name = 'air temperature at instrument level'
            temperature.units = 'K'
            temperature.standard_name = 'air_temperature'

            pressure = ancillary.createVariable('air_pressure', datatype='f8', dimensions=('time',))
            pressure.long_name = 'air pressure at instrument level'
            pressure.units = 'hPa'
            pressure.standard_name = 'air_pressure'

            # Create a separate group for each channel
            for channel_name, channel_parameters in channels.items():

                if channel_name not in list(self.channels.keys()):
                    raise ValueError('Channel name not one of {0}: {1}'.format(list(self.channels.keys()), channel_name))

                channel = self.channels[channel_name]

                group_name = "channel_{0}".format(channel_name.replace('.', '_'))  # Give channels groups a standard name

                g = f.createGroup(group_name)
                g.long_name = channel_parameters['long_name']
                g.detector_manufacturer = channel_parameters['detector_manufacturer']    # Optional
                g.detector_model = channel_parameters['detector_model']
                g.daq_manufacturer = channel_parameters['daq_manufacturer']
                g.daq_model = channel_parameters['daq_model']

                # Dimensions
                g.createDimension('profile', size=None)  # Infinite dimension
                g.createDimension('range', len(channel.z))

                # Variables
                name = g.createVariable('channel_id', 'c', dimensions=('name_strlen',))
                name.cf_role = 'timeseries_id'
                name.long_name = 'channel identification'

                laser_rep_rate = g.createVariable('laser_repetition_rate', 'f4')
                laser_rep_rate.long_name = 'nominal laser repetition rate'
                laser_rep_rate.units = 'Hz'

                emission_wavelength = g.createVariable('emission_wavelength', datatype='f8', )  # or dimensions=('profile',)
                emission_wavelength.long_name = 'emission wavelength'
                emission_wavelength.units = 'nm'
                emission_wavelength.comment = "could have dimension profile if measured."

                emission_energy = g.createVariable('emission_energy', datatype='f8', )  # or dimensions=('profile',)
                emission_energy.long_name = 'emission energy per pulse'
                emission_energy.units = 'mJ'
                emission_energy.comment = "could be scalar, if value is nominal."

                emission_pol = g.createVariable('emission_polarization', datatype='b')
                emission_pol.long_name = 'nominal emission poalrization'
                emission_pol.flag_values = '0b 1b 2b'
                emission_pol.flag_meanings = 'linear circular none'

                fov = g.createVariable('fov', datatype='f4')
                fov.long_name = 'channel field of view full angle'
                fov.units = 'mrad'
                fov.comment = 'simulated'

                detector_type = g.createVariable('detector_type', datatype='b')
                detector_type.long_name = 'detector type'
                detector_type.flag_values = '0b 1b'
                detector_type.flag_meanings = 'PMT APD'

                detection_mode = g.createVariable('detection_mode', datatype='b')
                detection_mode.long_name = 'detection mode'
                detection_mode.flag_values = '0b 1b'
                detection_mode.flag_meanings = 'analog photon_counting'

                detection_cw = g.createVariable('detection_wavelength', datatype='f8')
                detection_cw.long_name = 'center wavelength of detection filters'
                detection_cw.units = 'nm'
                detection_cw.standard_name = 'sensor_band_central_radiation_wavelength'

                detection_fwhm = g.createVariable('detection_fwhm', datatype='f8')
                detection_fwhm.long_name = 'FWHM of detection filters'
                detection_fwhm.units = 'nm'

                detection_pol = g.createVariable('detection_polarization', datatype='b')
                detection_pol.long_name = 'nominal detection poalrization'
                detection_pol.flag_values = '0b 1b 2b'
                detection_pol.flag_meanings = 'linear circular total'

                polarizer_angle = g.createVariable('polarizer_angle', datatype='f4', dimensions=('profile', ), zlib=True)
                polarizer_angle.long_name = 'polarizer angle in respect to laser plane of polarization'
                polarizer_angle.units = 'degree'
                polarizer_angle.comments = 'Optional'

                if channel.is_photon_counting:
                    dead_time_model = g.createVariable('dead_time_model', datatype='b')
                    dead_time_model.long_name = 'optimal dead time model of detection system'
                    dead_time_model.flag_values = '0b 1b 2b'
                    dead_time_model.flag_meanings = 'paralyzable non_paralyzable other'

                    dead_time = g.createVariable('dead_time', datatype='f8')
                    dead_time.long_name = 'dead time value'
                    dead_time.units = 'ns'
                    dead_time.comment = "Manufacturer. Source of the value."

                bin_length = g.createVariable('bin_length', datatype='f4')
                bin_length.long_name = "time duration of each bin"
                bin_length.units = 'ns'

                if channel.is_analog:
                    adc_bits = g.createVariable('adc_bits', datatype='i4')
                    adc_bits.long_name = 'analog-to-digital converter bits'
                    adc_bits.coordinates = "time"

                detector_voltage = g.createVariable('detector_voltage', datatype='f4', dimensions=('profile',), zlib=True)
                detector_voltage.long_name = 'detector voltage'
                detector_voltage.units = 'V'
                detector_voltage.coordinates = "time"

                if channel.is_photon_counting:
                    discriminator = g.createVariable('discriminator', datatype='f8', dimensions=('profile',))
                    discriminator.long_name = 'discriminator level'
                    discriminator.units = ''

                if channel.is_analog:
                    adc_range = g.createVariable('adc_range', datatype='f4', dimensions=('profile',),
                                                        zlib=True)
                    adc_range.long_name = 'analog-to-digital converter range'
                    adc_range.units = 'mV'
                    adc_range.coordinates = "time"

                pulses = g.createVariable('pulses', datatype='i4', dimensions=('profile',),
                                          zlib=True)
                pulses.long_name = 'accumulated laser pulses per record'
                pulses.coordinates = "time"

                nd_filter = g.createVariable('nd_filter_od', datatype='f8', dimensions=('profile',))
                nd_filter.long_name = "neutral density filter optical depth "
                nd_filter.coordinates = "time"

                trigger_delay = g.createVariable('trigger_delay', datatype='f4')
                trigger_delay.long_name = "channel trigger difference from pulse emission"
                trigger_delay.units = 'ns'
                trigger_delay.comments = 'Negative values for pre-trigger systems.'

                time = g.createVariable('time', datatype='f8', dimensions=('profile',),
                                        zlib=True)
                time.long_name = 'profile start time '
                time.units = "seconds since 1970-01-01 00:00:00"
                time.standard_name = "time"
                time.bounds = "time_bnds"
                time_bounds = g.createVariable('time_bnds', datatype='f8', dimensions=('profile', 'nv'), zlib=True)

                bin_time = g.createVariable('bin_time', datatype='f4', dimensions=('range',), zlib=True)
                bin_time.long_name = 'bin start time since channel trigger'
                bin_time.units = "ns"

                if channel.is_analog:
                    signal_units = 'mV'
                    signal_datatype = 'f8'
                else:
                    signal_units = 'counts'
                    signal_datatype = 'i8'

                signal = g.createVariable('signal', datatype=signal_datatype, dimensions=('profile', 'range'),
                                          zlib=True)
                signal.long_name = 'signal'
                signal.units = signal_units
                signal.coordinates = "time"
                signal.ancillary_variables = "signal_stddev"

                # If measured
                signal_stddev = g.createVariable('signal_stddev', datatype=signal_datatype, dimensions=('profile', 'range'),
                                          zlib=True)
                signal_stddev.long_name = 'signal standard deviation'
                signal_stddev.units = signal_units
                signal_stddev.coordinates = "time"
                signal_stddev.comments = "Only if measured. Should be removed if not."

                # Assign variables
                name[:len(channel_name)] = channel_name
                laser_rep_rate[:] = channel_parameters['laser_repetition_rate']
                emission_wavelength[:] = channel_parameters['emission_wavelength']
                emission_energy[:] = channel_parameters['emission_energy']
                emission_pol[:] = self._emission_pol_flag(channel_parameters['emission_polarization'])
                fov[:] = channel_parameters['fov']
                detector_type[:] = self._detector_type_flag(channel_parameters['detector_type'])
                detection_mode[:] = self._detection_mode_flag(channel_parameters['detection_mode'])
                detection_fwhm[:] = channel_parameters['filter_fwhm']
                detection_pol[:] = self._detection_pol_flag(channel_parameters['detection_polarization'])
                polarizer_angle[:] = channel_parameters['polarizer_angle'] * np.ones(len(channel.time))  # For now, assumed constant.

                if channel.is_photon_counting:
                    dead_time_model[:] = self._deadtime_model_flag(channel_parameters['dead_time_model'])
                    dead_time[:] = channel_parameters['dead_time']

                bin_length[:] = channel_parameters['bin_length']
                trigger_delay[:] = channel_parameters['trigger_delay']

                detector_voltage[:] = channel.hv

                if channel.is_analog:
                    adc_range[:] = channel.discriminator
                    adc_bits[:] = channel.adcbits
                else:
                    discriminator[:] = channel.discriminator

                pulses[:] = channel.laser_shots

                epoch = datetime.datetime(1970, 1, 1, tzinfo=pytz.utc)
                seconds_since_epoch = [(t - epoch).total_seconds() for t in channel.time]
                time[:] = seconds_since_epoch
                time_bounds[:, 0] = seconds_since_epoch
                time_bounds[:, 1] = seconds_since_epoch + channel.get_duration()

                bin_time[:] = channel.binwidth * np.arange(len(channel.z))

                signal[:] = channel.matrix

    def _deadtime_model_flag(self, model_str):
        """ Convert dead-time model string to byte flag.

        Parameters
        ----------
        model_str : str
           String describing the dead-time model (one of paralyzable, non-paralyzable, or other)

        Returns
        -------
        : int
           Byte encoding of dead-time model
        """
        choices = {'paralyzable': 0,
                   'non-paralyzable': 1,
                   'other': 2}

        if model_str not in list(choices.keys()):
            raise ValueError('Dead-time model is not one of {0}: {1}'.format(list(choices.keys()), model_str))

        return choices[model_str]

    def _detection_pol_flag(self, pol_str):
        """ Convert detection  polarization string to byte flag.

        Parameters
        ----------
        pol_str : str
           String describing the detection polarization (one of linear, circular, or total)

        Returns
        -------
        : int
           Byte encoding of detection polarization
        """
        choices = {'linear': 0,
                   'circular': 1,
                   'total': 2}

        if  pol_str not in list(choices.keys()):
            raise ValueError('Detection polarization is not one of {0}: {1}'.format(list(choices.keys()), pol_str))

        return choices[pol_str]

    def _detection_mode_flag(self, mode_str):
        """ Convert detection  mode string to byte flag.

        Parameters
        ----------
        mode_str : str
           String describing the detector mode (one of photon-counting or analog)

        Returns
        -------
        : int
           Byte encoding of detection mode
        """
        choices = {'analog': 0,
                   'photon-counting': 1,}

        if  mode_str not in list(choices.keys()):
            raise ValueError('Detection mode is not one of {0}: {1}'.format(list(choices.keys()), mode_str))

        return choices[mode_str]

    def _detector_type_flag(self, type_string):
        """ Convert emission string to byte flag.

        Parameters
        ----------
        type_string : str
           String describing the detector type (one of APD or PMT)

        Returns
        -------
        : int
           Byte encoding of detector type
        """
        choices = {'PMT': 0,
                   'APD': 1,}

        if  type_string not in list(choices.keys()):
            raise ValueError('Detector type is not one of {0}: {1}'.format(list(choices.keys()), type_string))

        return choices[type_string]

    def _emission_pol_flag(self, pol_string):
        """ Convert emission string to byte flag.

        Parameters
        ----------
        pol_string : str
           String describing the polarization (one of linear, circular, or none)

        Returns
        -------
        : int
           Byte encoding of polarization state
        """
        choices = {'linear': 0,
                   'circular': 1,
                   'none': 2}

        if pol_string not in list(choices.keys()):
            raise ValueError('Emission polarization not one of {0}: {1}'.format(list(choices.keys()), pol_string))

        return choices[pol_string]


class DivaLidarMeasurement(object):
    """ A class to read raw lidar files in DIVA format.

    Unlike other classes in this module, it does not inherit from BasicLidarMeasurement. This is done
    to avoid all the burden of backward compatibility. In the future this could be hosted also as a separte moduel.
    """

    def __init__(self, file_path, header_only=False):
        """
        This is run when creating a new object.

        Parameters
        ----------
        file_path : str
           Paths to the input netCDF file.
        header_only : bool
           If True, channel info are not loaded.
        """
        self.file_path = file_path
        self.file_name = os.path.basename(file_path)

        self.import_file(header_only)

    def import_file(self, header_only):
        """ Import data from a single DIVA file.
        """

        logger.debug('Importing file {0}'.format(self.file_name))

        self.channels = {}

        with netcdf.Dataset(self.file_path) as input_file:
            self.title = input_file.title
            self.source = input_file.source
            self.institution = input_file.institution
            self.references = input_file.references
            self.location = input_file.location
            self.data_version = input_file.data_version
            self.PI = input_file.PI
            self.PI_email = input_file.PI_email
            self.conversion_date_str = input_file.conversion_date
            self.conversion_date = datetime.datetime.strptime(input_file.conversion_date, '%Y-%d-%mT%H:%M:%SZ')
            self.comment = input_file.comment
            self.conventions = input_file.Conventions
            self.history = input_file.history

            self.latitude = input_file.variables['latitude'][:]
            self.longitude = input_file.variables['longitude'][:]

            self.lidar_zenith_angle = input_file.variables['lidar_zenith_angle'][:]
            self.lidar_azimuth_angle = input_file.variables['lidar_azimuth_angle'][:]
            self.lidar_altitude = input_file.variables['altitude'][:]

            ancillary = input_file.groups.pop('ancillary')

            self.meteo_time = ancillary.variables['time'][:]
            self.air_temperature_kelvin = ancillary.variable['air_temperature'][:]
            self.air_pressure_hpa = ancillary.variable['air_pressure'][:]

            self.available_channels = []
            for group_name, group in list(input_file.groups.items()):
                channel_name = group_name[8:]  # Remove 'channel_' prefix
                self.available_channels.append(channel_name)

                if not header_only:
                    self.channels[channel_name] = DivaChannel(channel_name, group)

    def import_channel(self, channel_name):
        """ Import a specific channel. """
        if channel_name not in self.available_channels:
            raise ValueError('Channel {0} not available. Should be one of {1}'.format(channel_name, self.available_channels))

        group_name = 'channel_{0}'.format(channel_name)

        with netcdf.Dataset(self.file_path) as input_file:
            group = input_file.groups[group_name]
            self.channels[channel_name] = DivaChannel(channel_name, group)


class DivaChannel(object):

    def __init__(self, channel_name, group):
        """ This is run when first creating the object.

        Parameters
        ----------
        channel_name : str
           Name of the group
        group : netCDF4.Group object
           An open netcdf group to initialize.
        """
        self.channel_name = channel_name

        self.long_name = group.long_name
        self.detector_manufacturer = getattr(group, 'detector_manufacturer', None)
        self.detector_model = getattr(group, 'detector_model', None)
        self.daq_manufacturer = getattr(group, 'daq_manufacturer', None)
        self.daq_model = getattr(group, 'daq_model', None)

        self.number_of_profiles = len(group.dimensions['profile'])
        self.number_of_bins = len(group.dimensions['range'])
        self.channel_id = group.variables['channel_id'][:]
        self.laser_repetition_rate = group.variables['laser_repetition_rate'][:]

        self.emission_energy_mJ = group.variables['emission_energy'][:]
        self.emission_polarization_flag = group.variables['emission_polarization'][:]
        self.emission_polarization = self._flag_to_polarization(self.emission_polarization_flag)
        self.field_of_view = group.variables['fov'][:]
        self.field_of_view_comment = group.variables['fov'].comment

        self.detector_type_flag = group.variables['detector_type'][:]
        self.detector_type = self._flag_to_detector_type(self.detector_type_flag)

        self.detection_mode_flag = group.variables['detection_mode'][:]
        self.detection_mode = self._flag_to_detector_type(self.detection_mode_flag)

        self.detection_wavelength_nm = group.variables['detection_wavelength'][:]
        self.detection_fwhm = group.variables['detection_fwhm'][:]

        self.detection_polarization_flag = group.variables['detection_polarization']
        self.detection_polariation = self._flag_to_detection_polarization(self.detection_polarization_flag)

        self.polarizer_angle_degrees = group.variables['polarizer_angle'][:]

        if self.is_photon_counting:
            self.dead_time_model_flag = group.variables['dead_time_model'][:]
            self.dead_time_model = self._flag_to_dead_time_model(self.dead_time_model_flag)

            self.dead_time = group.variables['dead_time'][:]
            self.dead_time_source = group.variables['dead_time'].comment
            self.discriminator = group.variables['discriminator'][:]

        if self.is_analog:
            self.adc_bits = group.variables['adc_bits'][:]
            self.adc_range = group.variables['adc_range'][:]

        self.bin_length_ns = group.variables['bin_length'][:]
        self.detector_voltage = group.variables['detector_voltage'][:]
        self.pulses = group.variables['pulses'][:]
        self.nd_filter_od = group.variables['nd_filter_od'][:]
        self.trigger_delay_ns = group.variables['trigger_delay'][:]
        self.time_since_epoch = group.variables['time'][:]

        self.time = [datetime.datetime.utcfromtimestamp(t) for t in self.time_since_epoch]
        self.bin_time_ns = group.variables['bin_time'][:]

        self.signal = group.variables['signal'][:]
        self.signal_units = group.variables['signal'].units

        signal_stddev_var = group.variables.pop('signal_stddev', None)

        if signal_stddev_var:
            self.signal_stddev = signal_stddev_var[:]
        else:
            self.signal_stddev = None

    def _flag_to_polarization(self, flag):
        """ Convert polarization flag to str"""
        if flag not in [0, 1, 2]:
            logger.warning('Polarization flag has unrecognized value: {0}'.format(flag))
            return ""

        values = {0: 'linear',
                  1: 'circular',
                  2: 'None'}

        return values[flag]

    def _flag_to_detector_type(self, flag):
        """ Convert detector type flag to str"""
        if flag not in [0, 1]:
            logger.warning('Detector type flag has unrecognized value: {0}'.format(flag))
            return ""

        values = {0: 'PMT',
                  1: 'APD'}

        return values[flag]

    def _flag_to_detection_mode(self, flag):
        """ Convert detector type flag to str"""
        if flag not in [0, 1]:
            logger.warning('Detection mode flag has unrecognized value: {0}'.format(flag))
            return ""

        values = {0: 'analog',
                  1: 'photon counting'}

        return values[flag]

    def _flag_to_detection_polarization(self, flag):
        """ Convert detector type flag to str"""
        if flag not in [0, 1, 2]:
            logger.warning('Detection polarization flag has unrecognized value: {0}'.format(flag))
            return ""

        values = {0: 'linear',
                  1: 'circular',
                  2: 'total'}

        return values[flag]

    def _flag_to_dead_time_model(self, flag):
        """ Convert detector type flag to str"""
        if flag not in [0, 1, 2]:
            logger.warning('Dead time model flag has unrecognized value: {0}'.format(flag))
            return ""

        values = {0: 'paralyzable',
                  1: 'non_paralyzable',
                  2: 'other'}

        return values[flag]

    @property
    def is_analog(self):
        return self.detection_mode_flag==0

    @property
    def is_photon_counting(self):
        return self.detection_mode_flag==1


from time import sleep
from pathlib import Path
from datetime import datetime, date, time
from typing import Any

# nc_convert imports
from gfatpy.lidar.utils.types import LidarName
from gfatpy.lidar.utils.utils import LIDAR_INFO
from gfatpy.lidar.nc_convert.utils import to_measurements
from gfatpy.lidar.nc_convert.converter import measurements_to_nc

# quicklook imports
from gfatpy.lidar.plot.quicklook import quicklook_from_file

# SCC imports
from loguru import logger
from gfatpy.lidar.scc import scc_access
from gfatpy.lidar.utils.types import MeasurementType
from gfatpy.lidar.nc_convert.types import Measurement
from gfatpy.lidar.scc.plot.scc_zip import SCC_zipfile
from gfatpy.utils.io import find_nearest_filepath, read_yaml
from gfatpy.lidar.scc.transfer import check_measurement_id_in_scc
from gfatpy.lidar.scc.licel2scc import licel2scc, licel2scc_depol
from gfatpy import GFATPY_DIR

RAW_DIR = Path(r"tests\datos\RAW")
PRODUCTS_DIR = Path(r"tests\datos\PRODUCTS")


def nc_convert(lidar_name: str, target_date: str | date | None = None) -> str:
    lidar = LidarName(lidar_name.lower())

    if target_date is None:
        target_date = datetime.now().date()
    elif isinstance(target_date, str):
        target_date = datetime.strptime(target_date, "%Y-%m-%d").date()

    raw_dir = (
        RAW_DIR
        / lidar.value
        / f"{target_date.year}"
        / f"{target_date.month:02}"
        / f"{target_date.day:02}"
    )

    if not raw_dir.exists():
        return f"{raw_dir} not found in {RAW_DIR}."

    measurements = to_measurements(raw_dir.glob("RS*"))
    measurements_to_nc(
        measurements,
        lidar_name=lidar,
        raw_dir=RAW_DIR,
        output_dir=PRODUCTS_DIR,
    )
    netcdf_dir = (
        PRODUCTS_DIR
        / lidar
        / "1a"
        / f"{target_date.year}"
        / f"{target_date.month:02}"
        / f"{target_date.day:02}"
    )

    return f"nc files created in {netcdf_dir.absolute().as_posix()}."


def quicklook(
    lidar_name: str, channel: str, target_date: str | date | None = None
) -> str:

    if target_date is None:
        target_date = datetime.now().date()
    elif isinstance(target_date, str):
        target_date = datetime.strptime(target_date, "%Y-%m-%d").date()
    datestr = target_date.strftime("%Y%m%d")
    year, month, day = target_date.year, target_date.month, target_date.day

    data_dir = (
        PRODUCTS_DIR / lidar_name / "1a" / f"{year:04}" / f"{month:02}" / f"{day:02}"
    )
    lidarnick = LIDAR_INFO["metadata"]["name2nick"][lidar_name]
    files = [*data_dir.glob(f"{lidarnick}_1a_Prs_rs_*{year:04}{month:02}{day:02}.nc")]

    if not files:
        return f"No files found in {data_dir}."
    if len(list(files)) > 1:
        return f"More than one RS lidar netcdf file found in {data_dir}."

    quicklook_dir = PRODUCTS_DIR / lidar_name / "quicklooks"
    quicklook_dir.mkdir(parents=True, exist_ok=True)

    figures = []
    for file_ in files:
        logger.info(f"{file_.name} measurements found in {data_dir}.")

        quicklook_from_file(
            filepath=file_, channels=[channel], output_dir=quicklook_dir
        )

        figures.append(
            PRODUCTS_DIR
            / f"{lidar_name}"
            / "quicklooks"
            / f"quicklook_{lidarnick}_{channel}_{datestr}.png"
        )
    return f"{figures[0]} created."


def convert_scc(
    lidar_name: str,
    scc_id: int,
    temperature: float = 20.0,
    pressure: float = 1013.25,
    target_date: str | date | None = None,
    intervals: list[Any] = [
        (time(3, 15), time(3, 45)),
        (time(3, 45), time(4, 15)),
    ],
    **kwargs,
) -> list[str] | str:

    if target_date is None:
        target_date = datetime.now().date()
    elif isinstance(target_date, str):
        target_date = datetime.strptime(target_date, "%Y-%m-%d").date()

    logger.info(f"Launching SCC conversion for {lidar_name} on {target_date}.")

    year, month, day = target_date.year, target_date.month, target_date.day

    nc_scc_files = []
    for ini_hour, end_hour in intervals:

        measurement_id = f"{year:04}{month:02}{day:02}gra{ini_hour.strftime('%H%M')}"
        target_period = (
            datetime.combine(target_date, ini_hour),
            datetime.combine(target_date, end_hour),
        )

        nc_scc_dir = (
            PRODUCTS_DIR
            / f"{lidar_name}"
            / "scc"
            / f"scc{scc_id}"
            / f"{year:04}"
            / f"{month:02}"
            / f"{day:02}"
        )
        nc_scc_dir.mkdir(parents=True, exist_ok=True)

        scc_config_dir = kwargs.get(
            "scc_config_dir", GFATPY_DIR / "lidar" / "scc" / "scc_configFiles"
        )
        breakpoint()
        scc_config_fn = find_nearest_filepath(
            scc_config_dir,
            f"alh_parameters_scc_{scc_id}_*.py",
            4,
            target_period[0],
            and_previous=True,
        )
        logger.info(f"Using {scc_config_fn.name} for SCC conversion.")

        # Define custom class for regular lidar measurements
        CustomLidarMeasurement = licel2scc.create_custom_class(
            scc_config_fn.absolute().as_posix(),
            use_id_as_name=True,
            temperature=temperature,
            pressure=pressure,
        )

        data_dir = (
            RAW_DIR
            / "alhambra"
            / f"{target_date.year:04}"
            / f"{target_date.month:02}"
            / f"{target_date.day:02}"
        )
        measurements = to_measurements(
            data_dir.glob(
                f"RS_{target_date.year:04}{target_date.month:02}{target_date.day:02}*"
            )
        )
        if not measurements:
            logger.warning(f"No measurements found in {data_dir}.")
            continue

        logger.info(f"{len(measurements)} measurements found in {data_dir}.")
        logger.info("First measurement:")
        logger.info(print(measurements[0]))

        file_set = set()
        for measurement in measurements:
            file_set = measurement.get_filepaths(within_period=target_period)
            if len(file_set) == 0:
                continue
            else:
                logger.info(f"{len(file_set)} files found in {measurement.path.name}.")
                break

        if len(file_set) < 30:
            logger.warning(
                f"Interval {ini_hour.strftime('%H:%M')}-{end_hour.strftime('%H:%M')} has less than 30 files."
            )
            continue

        rs_files_slot = sorted([file_.as_posix() for file_ in file_set])

        dc_path = measurement.path.parent / measurement.path.name.replace("RS", "DC")
        if dc_path.exists():
            dc_measurement = Measurement(
                type=MeasurementType("DC"),
                path=dc_path,
            )
            logger.info(f"Found coincident DC measurement: {dc_measurement.path.name}.")
        else:
            try:
                dc_measurement = to_measurements(data_dir.glob("DC*"))[0]
                logger.info(
                    f"Found non-coincident DC measurement: {dc_measurement.path.name}."
                )
            except IndexError:
                logger.warning(f"No DC measurements found in {data_dir}.")
                continue

        if dc_measurement.is_zip:
            dc_measurement.extract_zip()
            if dc_measurement.unzipped_path is not None:
                dc_files_patt = (
                    dc_measurement.unzipped_path / "**" / "*.[0-9]*"
                ).as_posix()
            else:
                raise ValueError(f"Error extracting {dc_measurement.path}.")
        else:
            dc_files_patt = dc_measurement.path.as_posix()

        licel2scc.convert_to_scc(
            CustomLidarMeasurement,
            rs_files_slot,
            dc_files_patt,
            measurement_id,
            output_dir=nc_scc_dir,
        )
        if not (nc_scc_dir / f"{measurement_id}.nc").exists():
            logger.warning(f"Error converting {measurement_id} to SCC.")
            continue
        else:
            logger.info(f"{measurement_id}.nc created.")
            nc_scc_files.append(nc_scc_dir / f"{measurement_id}.nc")
    if nc_scc_files == []:
        return f"No SCC files created for {lidar_name} on {target_date}."
    return [nc.name for nc in nc_scc_files]


def convert_scc_dp(
    lidar_name: str,
    scc_id: int,
    temperature: float = 20.0,
    pressure: float = 1013.25,
    target_date: str | date | None = None,
    licel_timezone: str = "UTC",
    **kwargs,
) -> list[str] | str:

    if target_date is None:
        target_date = datetime.now().date()
    elif isinstance(target_date, str):
        target_date = datetime.strptime(target_date, "%Y-%m-%d").date()
    year, month, day = target_date.year, target_date.month, target_date.day
    date_str = target_date.strftime("%Y%m%d")

    scc_config_dir = kwargs.get(
        "scc_config_dir", GFATPY_DIR / "lidar" / "scc" / "scc_configFiles"
    )
    scc_config_fn = find_nearest_filepath(
        scc_config_dir,
        f"alh_parameters_scc_{scc_id}_*.py",
        4,
        datetime.combine(target_date, time(0, 0)),
        and_previous=True,
    )
    logger.info(f"Using {scc_config_fn.name} for SCC conversion.")

    CustomLidarMeasurement = licel2scc_depol.create_custom_class(
        scc_config_fn.absolute().as_posix(),
        use_id_as_name=True,
        temperature=temperature,
        pressure=pressure,
    )

    # Find DP measurements
    data_dir = RAW_DIR / lidar_name / f"{year}" / f"{month:02}" / f"{day:02}"
    measurements = to_measurements(data_dir.glob(f"DP_{date_str}*.zip"))
    if not measurements:
        logger.warning(f"No measurements found in {data_dir}.")
        return f"No measurements found in {data_dir}."
    logger.info(f"{len(measurements)} measurements found in {data_dir}.")
    output_paths = []
    for measurement in measurements:
        if measurement.is_zip:
            measurement.extract_zip()
            if measurement.unzipped_path is not None:
                plus45_files = [
                    file_.as_posix()
                    for file_ in measurement.unzipped_path.rglob(f"{date_str}/+45/*.*")
                ]
                minus45_files = [
                    file_.as_posix()
                    for file_ in measurement.unzipped_path.rglob(f"{date_str}/-45/*.*")
                ]
            else:
                raise ValueError(f"Error extracting {measurement.path}.")
        else:
            plus45_files = [
                file_.as_posix()
                for file_ in measurement.path.rglob(f"{date_str}/+45/*.*")
            ]
            minus45_files = [
                file_.as_posix()
                for file_ in measurement.path.rglob(f"{date_str}/-45/*.*")
            ]

        if plus45_files == []:
            raise Exception("No P45 files found")

        if plus45_files == []:
            raise Exception("No N45 files found")

        if measurement.has_dc:
            dc_tmp_path = measurement.dc_path
            if dc_tmp_path is not None:
                dc_measurement = to_measurements(data_dir.glob(dc_tmp_path.name))[0]
                if dc_measurement.is_zip:
                    dc_measurement.extract_zip()

                    if dc_measurement.unzipped_path is not None:
                        dark_files = [
                            file_.as_posix()
                            for file_ in dc_measurement.unzipped_path.rglob(
                                f"{date_str}/[RD]*.*"
                            )
                        ]
                    else:
                        raise ValueError(f"Error extracting {dc_measurement.path}.")
                else:
                    dark_files = [
                        file_.as_posix()
                        for file_ in dc_measurement.path.rglob(f"{date_str}/[RD]*.*")
                    ]
        hour_str = measurement.path.name.split(".")[0].split("_")[-1]
        measurement_id = f"{year:04}{month:02}{day:02}gra{hour_str}"
        output_path = (
            PRODUCTS_DIR
            / lidar_name
            / "scc"
            / f"scc{scc_id}"
            / f"{year}"
            / f"{month:02}"
            / f"{day:02}"
            / f"{measurement_id}.nc"
        )

        if not output_path.parent.exists():
            output_path.parent.mkdir(parents=True, exist_ok=True)

        CustomLidarMeasurement = licel2scc_depol.create_custom_class(
            scc_config_fn, True, temperature, pressure, licel_timezone
        )

        CustomDarkMeasurement = licel2scc_depol.create_custom_dark_class(
            scc_config_fn, True, temperature, pressure, licel_timezone
        )

        measurement = CustomLidarMeasurement(plus45_files, minus45_files)

        if dark_files:
            measurement.dark_measurement = CustomDarkMeasurement(dark_files)  # type: ignore
        else:
            raise Exception("No dark measurement files found.")

        try:
            measurement = measurement.subset_by_scc_channels()
        except ValueError as err:
            raise err

        # Save the netcdf
        measurement.set_measurement_id(output_path.name.split(".")[0])
        measurement.save_as_SCC_netcdf(output_dir=output_path.parent)
        output_paths.append(output_path.absolute().as_posix())
    return output_paths


def send_to_scc(
    lidar_name: str,
    scc_id: int,
    info_scc_config_name: str,
    target_date: str | date | None = None,
    **kwargs,
) -> list[str] | str:

    # Check if the environment variable is set
    info_scc_dir = Path(
        kwargs.get("info_scc_dir", (GFATPY_DIR / "env_files").absolute().as_posix())
    )
    config_path = info_scc_dir / info_scc_config_name

    if not config_path.exists():
        raise ValueError(
            "The environment variable 'SCC_CONFIG_FILE' is not set in the .env"
        )

    logger.info(f"Using {config_path} for SCC conversion.")

    # Construct the file path
    SCC_INFO = read_yaml(config_path)
    SCC_SERVER_SETTINGS = SCC_INFO["server_settings"]

    if target_date is None:
        target_date = datetime.now().date()
    elif isinstance(target_date, str):
        target_date = datetime.strptime(target_date, "%Y-%m-%d").date()

    year, month, day = target_date.year, target_date.month, target_date.day

    scc_dir = (
        PRODUCTS_DIR
        / lidar_name
        / "scc"
        / f"scc{scc_id}"
        / f"{year:04}"
        / f"{month:02}"
        / f"{day:02}"
    )
    files = scc_dir.glob("*.nc")
    succeeded_upload_files = []
    for file_ in files:
        logger.info(f"Launching SCC conversion for {lidar_name} on {target_date}.")
        measurement_id = file_.name.replace(".nc", "")

        scc_obj = scc_access.SCC(
            tuple(SCC_SERVER_SETTINGS["basic_credentials"]),
            SCC_SERVER_SETTINGS["output_dir"],
            SCC_SERVER_SETTINGS["base_url"],
        )

        measurement_exists, _ = check_measurement_id_in_scc(
            SCC_SERVER_SETTINGS, measurement_id
        )

        if measurement_exists:
            logger.info(
                f"Measurement {measurement_id} exists in SCC: {measurement_exists}."
            )
            return f"{measurement_id} already exists in SCC."

        scc_obj.login(SCC_SERVER_SETTINGS["website_credentials"])
        measurement_id_from_server = scc_obj.upload_file(
            filename=file_, system_id=scc_id
        )

        scc_obj.logout()
        if not measurement_id_from_server:
            logger.warning(f"Error uploading {measurement_id}.nc to SCC.")
            continue

        product_scc_dir = scc_dir / "products"
        product_scc_dir.mkdir(parents=True, exist_ok=True)
        SCC_SERVER_SETTINGS["output_dir"] = product_scc_dir

        scc_obj = scc_access.SCC(
            tuple(SCC_SERVER_SETTINGS["basic_credentials"]),
            SCC_SERVER_SETTINGS["output_dir"],
            SCC_SERVER_SETTINGS["base_url"],
        )

        scc_obj.login(SCC_SERVER_SETTINGS["website_credentials"])

        # Manejo de reintentos
        retries = 5
        for attempt in range(retries):
            try:
                scc_obj.monitor_processing(measurement_id)
                succeeded_upload_files.append(file_.name)
                break
            except TimeoutError:
                if attempt < retries - 1:
                    # Esperar un tiempo antes de reintentar
                    sleep(60)
                else:
                    logger.warning(
                        f"Process was queued too long after {retries} attempts."
                    )
                    continue

        scc_obj.logout()
    return succeeded_upload_files


def plot_scc(
    lidar_name: str, scc_id: int, target_date: str | date | None = None
) -> list[str] | str:
    def flatten_list(nested_list):
        flattened = []
        for item in nested_list:
            if isinstance(item, list):
                flattened.extend(flatten_list(item))
            else:
                flattened.append(item)
        return flattened

    if target_date is None:
        target_date = datetime.now().date()
    elif isinstance(target_date, str):
        target_date = datetime.strptime(target_date, "%Y-%m-%d").date()
    year, month, day = target_date.year, target_date.month, target_date.day

    scc_dir = (
        PRODUCTS_DIR
        / lidar_name
        / "scc"
        / f"scc{scc_id}"
        / f"{year:04}"
        / f"{month:02}"
        / f"{day:02}"
        / "products"
    )

    if not scc_dir.exists():
        return f"{scc_dir} does not exist."

    logger.info(f"Searching for products in {scc_dir}.")
    products = [*scc_dir.glob("*.zip")]
    logger.info(f"{len(products)} products found in {scc_dir}.")

    # Plotting scc results
    plot_scc_dir = scc_dir / "plots"
    plot_scc_dir.mkdir(parents=True, exist_ok=True)

    if not products:
        return f"No SCC files found in {scc_dir}."
    plot_paths = []
    for product in products:
        logger.info(f"Plotting {product.name}.")
        try:
            scc_zip = SCC_zipfile(product)
        except Exception as e:
            logger.warning(f"Error reading {product} as SCC zipfile.")
            continue
        plot_paths_ = scc_zip.plot(
            output_dir=plot_scc_dir, dpi=150, range_limits=(0, 10)
        )
        if isinstance(plot_paths_, list):
            plot_paths.extend(plot_paths_)
        else:
            plot_paths.append(plot_paths_)

    flattened_list = flatten_list(plot_paths)

    plot_names = [plot.name for plot in flattened_list]
    return plot_names
